/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "boundary.h"

//Override of terminal_vertex virtual function
Vertex Boundary::terminal_vertex() {
	return global::End_Boundary;
}

//Sort algorithm (by successive rotations)
void Boundary::sort() {
	//sort is useless when boundary is shorter than 3 characters (at least "xy." is needed)
	int length=r_vtx.size();	//boundary length
	if (length<3) {
		return;
	}
	
	//delete 'end of boundary' character (rotation is easier)
	r_vtx.pop_back();
	
	//temporary object containing the boundary's rotations
	Line rotations=r_vtx;
	
	//iterators used by rotate function
	Line::iterator p, q, r;
	p=rotations.begin();
	r=rotations.begin(); r++; //rotate character by character on the left
	q=rotations.end();
	
	//look for the best rotation (r_vtx.size()-1 possible rotations, the last one is unnecessary)
	//r_vtx is always equal to the best rotation found
	int i;
	for(i=1;i<length-1;i++) { 
		rotate(p,r,q);		//rotate one character
		if (rotations<r_vtx) {	//check whether this new rotation is the best
			r_vtx=rotations;
		}
	}
	
	//add the 'end of boundary' character
	r_vtx.push_back(global::End_Boundary);
}

//Reverse boundary's orientation
void Boundary::reverse_orientation() {
	//delete 'end of boundary' character
	r_vtx.pop_back();
	
	//reverse orientation with STL reverse algorithm
	r_vtx.reverse();
	
	//add back 'end of boundary' character
	r_vtx.push_back(global::End_Boundary);
}

// return true if the boundary is a single unused spot : 0.
bool Boundary::isSingleSpot() const {
	if (r_vtx.size()==2) {
		if (r_vtx.front()==0) {
			return true;
		}
	}
	
	return false;
}


//-------------------------------------------------------------------------------------------------
// Boundary's children computation
//-------------------------------------------------------------------------------------------------

// 1. Boundary's children when linked to itself
//---------------------------------------------
//this function implements rules of linking (valid moves or not)
//a child is a pair of boundaries : one in the inner domain and one in the outer domain
void Boundary::compute_oneBoundary_children() {
  //note : boundary should not be empty (because of canonization)
  //we treat differently the single vertex boundaries (0. 1. or  2.) and the other ones
	children_couple.clear();	//children set
	
	if (r_vtx.size()<2) {
		Error(QString("Boundary::compute_oneBoundary_children : Boundary is empty"), false);
		return;
	}
	
	Line bnd;		//temporary boundary object for successive rotations
	couple couple_bnd;	//pair of boundaries : inner+outer
	
	//----------Boundary is a single vertex----------------------------------------------------
	if (r_vtx.size()==2) {
		//only possible move : link the single vertex to itself
		//which is a valid move only if vertex degree is 0 or 1
		Vertex v=r_vtx.front();
		if (v==0 || v==1) {
			//boundary in the outer domain
			if(v==0) {bnd.push_back(global::Child_vtx_1);}	//vertex "0" is renamed "Child_vtx_1"
			if(v==1) {bnd.push_back(3);}				//vertex "1" is renamed "3"
			bnd.push_back(global::Child_vtx_3);			//creation of a new vertex
			bnd.push_back(global::End_Boundary);			//'end of boundary'
			
			//stock result
			couple_bnd.first=bnd;
			  //boundary in the inner domain : same as in the outer one
			couple_bnd.second=bnd;
			children_couple.insert(couple_bnd);
		}
		return;
	}
	
	//----------Boundary has more than one vertex----------------------------------------------
	r_vtx.pop_back();	//delete 'end of boundary' character (we'll push it back later)
	
	Line::iterator s1, s2;	//vertex iterator
	Vertex v1, v2;	//vertex value
	
	//double loop that try to link all possible couple of vertices in the boundary
	for(s1=r_vtx.begin(); s1!=r_vtx.end(); s1++) {
	for(s2=s1; s2!=r_vtx.end(); s2++) {
		//test if move is possible
		//forbidden links : a vertex "2" to itself or a letter to the same letter
		if( !(s1==s2 && *s1==2) && !(*s1==*s2 && *s1>2) ) {
			v1=*s1; v2=*s2;	//save vertices name
			
			//-----give a real name to generic vertices------------------------------
			//we need to give generic vertices a real name when they are used in a link
			if(s1!=s2) {
				//link two different vertices
				if(v1==1) {*s1=global::Child_vtx_1;}	//a vertex of degree 1 becomes a letter
				if(v2==1) {*s2=global::Child_vtx_2;}	//we need different letters for s1 and s2
				//vertices of degree 2 -> degree 3 (they'll be erased on canonization step)
				if(v1==2) {*s1=3;}
				if(v2==2) {*s2=3;}
			} else {
				//link a vertex to itself (possible only if its degree is 1)
				//vertex of degree 1 -> degree 3
				if(v1==1) {*s1=3;}
			}
			
			//-----boundary's child in the outer domain------------------------------
			bnd.clear();
			bnd.insert(bnd.end(), r_vtx.begin(), s1);	//caution : s1 is not inserted yet
			bnd.push_back(*s1);
			bnd.push_back(global::Child_vtx_3);		//add a new vertex (with an available name)
			bnd.insert(bnd.end(), s2, r_vtx.end());
			bnd.push_back(global::End_Boundary);		//end of boundary
			couple_bnd.first=bnd;
			
			//-----boundary's child in the inner domain------------------------------
			bnd.clear();
			bnd.insert(bnd.end(), s1, s2);		//caution : s2 is not inserted yet
			bnd.push_back(*s2);			//add s2 in all cases, we don't care that s1!=s2 or s1==s2
			bnd.push_back(global::Child_vtx_3);	//add a new vertex
			bnd.push_back(global::End_Boundary);
			couple_bnd.second=bnd;
			
			//-----add the pair (outer-inner) to the list of children----------------
			children_couple.insert(couple_bnd);
			
			*s1=v1; *s2=v2;	//give back to vertices their correct names
		}
	}}
	
	r_vtx.push_back(global::End_Boundary);	//put back the 'end of boundary' character
}

// 2. Boundary's children when linked to another boundary
//-------------------------------------------------------
//there's no linking rule in this function (rules are enforced in Domain::compute_twoBoundaries_children)
//algorithm consists mainly in rotation (complete theory on separate papers)
void Boundary::compute_twoBoundaries_children() {
  //note 1: boundary should not be empty (canonization should have been used)
  //note 2: children are not boundaries but boundary parts, ready to be linked with other boundary parts
  //        this is why children don't have any 'end of boundary' character
	children.clear();
	
	if (r_vtx.size()<=1) {
		Error(QString("Boundary::compute_twoBoundaries_children : Boundary is empty"), false);
		return;
	}
	
	//----------compute children-----------------------------------------------------
	Line bnd=r_vtx;		//temporary object for successive rotations
	bnd.pop_back();		//delete 'end of boundary' character
	Line::iterator p, q, r;
	Vertex v;		//first character of a particular rotation
	int i;
	int length=bnd.size();	//there are "length" different children
	
	for(i=0; i<length; i++) {
		v=bnd.front();
		//in case of a generic vertex, we need to rename it
		switch(v) {
		case 0:
			//0->1
			bnd.pop_front();
			bnd.push_front(1);
			children.insert(bnd);
			bnd.pop_front();
			bnd.push_front(0);
			break;
		case 1:
			//1->Child_vtx_1(reserved name)
			//Child_vtx_1 is renamed in Child_vtx_2 in Domain::compute_twoBoundaries_children when necessary
			bnd.pop_front();
			bnd.push_front(global::Child_vtx_1);
			bnd.push_back(global::Child_vtx_1);
			children.insert(bnd);
			bnd.pop_front();
			bnd.pop_back();
			//add back the deleted 1
			bnd.push_front(1);
			break;
		case 2:
			//2->3, we just delete it
			bnd.pop_front();
			children.insert(bnd);
			bnd.push_front(2);
			break;
		default:
			//in case of a "normal" vertex we add the vertex at the end of boundary
			//it will disappear with its other occurrences at the next canonization
			bnd.push_back(v);
			//add the obtained boundary to the children list
			children.insert(bnd);
			//delete from the boundary the added vertex
			bnd.pop_back();
		}
		
		//rotation of bnd object
		p=bnd.begin();
		r=bnd.begin(); r++; //rotation of one vertex only to the left
		q=bnd.end();
		rotate(p,r,q);
	}
}


//-------------------------------------------------------------------------
// lower-upper case distinction :
// these functions renames appropriate upper-case vertices in lower-case
// (appears two times in the same boundary)
//--------------------------------------------------------------------------

//Rename upper-case and temporary vertices in lower-case ones (when needed)
void Boundary::distinguish_low_up() {
  //step 1 : find biggest used lower-case letter
	Line::iterator s;
	Vertex next_letter=global::First_low_vtx; //first unused lower-case Vertex
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		//test lower-case and bigger of current first unused lower-case letter
		if(is_lower_case(*s) && *s>=next_letter) {
			next_letter=*s+1; //update first unused lower-case Vertex (*s is used)
		}
	}
	
  //step 2 : rename upper-case letters in lower-case ones (when needed)
	Line::iterator p, q;
	p=r_vtx.begin();
	q=r_vtx.end();
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		//test if Vertex is a temp/upper-case one + appears two times in the boundary (->use lower-case in that case)
		if(is_upper_case(*s) || is_temp_vtx(*s)) {
			if(count(p,q,*s)==2) {
				//WARNING ! for an unknown reason we need to precise explicitly Vertex type in the replace function
				replace(p, q, (Vertex) *s, (Vertex) next_letter); //rename
				next_letter++; //update first unused lower-case Vertex
			}
		}
	}
	
  //security on the number of used letters
	next_letter--; //last used lower-case letter
	if (next_letter>global::Last_low_vtx) {
		Error(QString("Boundary::distinguish_low_up : lower-case letters capacity overflow"), true);
	}
}

//rename lower-case letters in their order of appearance
void Boundary::rename_low_in_order() {
	vector<int> new_names(global::Last_vtx+1);  //Vertices new names
	
	//initialization, convention : 0 means here "new name is not yet decided"
	fill(new_names.begin(), new_names.end(), 0);
	
	Line::iterator s;
	//give new names (begins at First_low_vtx)
	int next_letter; //first unused Vertex name
	
	next_letter=global::First_low_vtx;
	for(s= r_vtx.begin(); s != r_vtx.end(); s++) {
		//test if Vertex is a lower-case letter
		if(is_lower_case(*s)) {
			//test if a new name already exists
			if (new_names[*s]==0) {
				//new name doesn't exist : give it one
				new_names[*s]=next_letter;
				*s=next_letter;
				next_letter++;
			} else {
				//use already known new name
				*s=new_names[*s];
			}
		}
	}
	
	//security on the number of used letters (probably useless, the useful one
	//is the security in Boundary::distinguish_low_up())
	next_letter--; //last used lower-case letter
	if (next_letter>global::Last_low_vtx) {
		Error(QString("Boundary::rename_low_in_order : lower-case letters capacity overflow"), true);
	}
}

//avoid lower-case letters : use upper-case ones instead
void Boundary::lower_to_upper() {
	Line::iterator s, p, q;
	p=r_vtx.begin();
	q=r_vtx.end();
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		//if Vertex is a lower-case letter, rename it with the first available upper-case letter
		if(is_lower_case(*s)) {
			//WARNING ! for an unknown reason we need to precise explicitly Vertex type in the replace function
			replace(p, q, (Vertex) *s, (Vertex) Representation::free_upper_vtx); //rename
			Representation::free_upper_vtx++; //update first unused upper-case letter
		}
	}
	
	//security on the number of used uper-case letters
	if ( (free_upper_vtx-1) > global::Last_up_vtx) {
		Error(QString("Boundary::lower_to_upper : upper-case letters capacity overflow"), true);
	}
}
