/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "domain.h"

//override terminal_vertex virtual function
Vertex Domain::terminal_vertex() {
	return global::End_Domain;
}

//-------------------------------------------------------------------------------------------------
// Reduction/canonization functions
//-------------------------------------------------------------------------------------------------

//sort each boundary, and then the list of boundaries
void Domain::simple_sort() {
	//sort each boundary
	list<Boundary>::iterator Bi;
	for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
		Bi->sort();
	}
	
	//sort the list of boundaries
	r_str.sort();
}

//sort domain (with possible orientation inversion)
void Domain::sort() {
	update_representation();
	
	//----------normal orientation---------------------
	change_structure();
		simple_sort();        //sort boundaries
	update_representation();
	
	Line r_sav=r_vtx;         //save the result
	
	//----------inverse orientation--------------------
	//reverse each boundary
	change_structure();
		list<Boundary>::iterator Bi;
		for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
			Bi->reverse_orientation();
		}
		
		simple_sort();       //sort boundaries
	update_representation();
	
	//----------compare results------------------------
	//keep the best orientation
	if (r_sav<r_vtx) {
		change_vertex();
		r_vtx=r_sav;
	}
}

//returns true when the Domain is dead (less than 2 lives)
bool Domain::dead() const {
  //note : this algorithm counts too much lives (when the same letter appears more than one time)
  //but this has no influence at all on the function, because when this function is called,
  //there is no more dead domain where the same letter appears more than once
	Line::const_iterator s;
	int lives=0;
	
	//this loop stops as soon as the domain is proved alive
	for(s=r_vtx.begin(); s != r_vtx.end() && lives<2; s++) {
		if(is_letter_vtx(*s)) {
			//normal letter : one life
			lives += 1;
		} else
		if(*s<=2) {
			//generic vertex : (3-*s) lives
			lives += (3-*s);
		}
	}
	
	//returns dead/alive state (true if less than 2 lives)
	return (lives<2);
}

//Removes all the End_Boundary characters (only the last one remains) in domains with at most 3 lives.
void Domain::merge_boundaries() {
  //note : this algorithm counts too much lives (when the same letter appears more than one time)
  //but this has no influence at all on the function
	Line::const_iterator s;
	int lives=0;		//number of lives
	
	//this loop stops as soon as the domain is found with more than 3 lives
	for(s=r_vtx.begin(); s != r_vtx.end() && lives<4; s++) {
		if(is_letter_vtx(*s)) {
			//normal letter : one life
			lives += 1;
		} else
		if(*s<=2) {
			//generic vertex : (3-*s) lives
			lives += (3-*s);
		}
	}
	
	if(lives<4 && lives>1){
		change_vertex();
		
		//remove all End_Boundary characters
		r_vtx.remove(global::End_Boundary);
		
		//remove the End_Domain character
		r_vtx.pop_back();
		
		//add End_Boundary and End_Domain
		r_vtx.push_back(global::End_Boundary);
		r_vtx.push_back(global::End_Domain);
	}
}


//-------------------------------------------------------------------------------------------------
//Children computation
//-------------------------------------------------------------------------------------------------

//there are two different kinds of children :
//those obtained by linking a boundary to itself
//and those obtained by linking two different boundaries
void Domain::compute_children() {
	update_representation();
	
	//if there are lower case letters, rename it in upper case ones (necessary to compute children)
	list<Boundary> sav;
	if (exist_lower(r_vtx)) {
		sav=r_str;
		lower_to_upper();
	}
	
	children.clear();
	
	//children obtained by linking a boundary to itself
	compute_oneBoundary_children();
	
	//children obtained by linking two different boundaries
	compute_twoBoundaries_children();
	
	//get back (if necessary) the original list of boundaries with lower-case letters
	if(!sav.empty()) {
		r_str=sav;
	}
}

//Children of a domain when we link a boundary to itself
void Domain::compute_oneBoundary_children() {
	list<Boundary>::iterator Bi, Bk;
	set<couple>::const_iterator Ci, Rj;	//children and partition of Bi
	
	list<Boundary> unused_bnd;	//list of "unused" boundaries
	Line child;			//one child of the domain
	
	//loop on the boundary list
	for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
		Bi->compute_oneBoundary_children();	//children of Bi (a child is a pair of boundaries)
		
		//--- a) create list of unused boundaries
		unused_bnd.clear();
		for(Bk=r_str.begin(); Bk!=r_str.end(); Bk++) {
			if(Bk!=Bi) {
				//this is an unused boundary -> we add it to unused_bnd
				unused_bnd.push_back(*Bk);
			}
		}
		
		//--- b) create set of all possible partitions in two blocks
		partition(unused_bnd);
		
		//--- c) create children of the domain
		//there is one child for each child of Bi and each partition of unused boundaries
		for(Ci=Bi->children_couple.begin(); Ci!=Bi->children_couple.end(); Ci++) {
		for(Rj=possible_partition.begin(); Rj!=possible_partition.end(); Rj++) {
			child.clear();
			
			child.insert(child.end(),Ci->first.begin(),Ci->first.end());   //copy the "outer boundary"
			child.insert(child.end(),Rj->first.begin(),Rj->first.end());   //copy the partition first block
			child.push_back(global::End_Domain);                           //add 'end of domain' character
			
			child.insert(child.end(),Ci->second.begin(),Ci->second.end()); //copy the "inner boundary"
			child.insert(child.end(),Rj->second.begin(),Rj->second.end()); //copy the partition second block
			child.push_back(global::End_Domain);                           //add 'end of domain' character
			
			//add the child to the domain's list of children
			children.insert(child);
		}}
	}
}

//Children of a domain when we link two different boundaries
void Domain::compute_twoBoundaries_children() {
  //note: children are added to children
	list<Boundary>::iterator Bi, Bj, Bk;	//boundaries list iterators
	set<Line>::iterator Ci, Cj;		//children set iterators
	
	Line unused_bnd;	//vertex list of unused boundaries
	Line child;		//domain's child
	
	//for each boundary, compute children (boundary bribes)
	for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
		Bi->compute_twoBoundaries_children();
	}
	
	//double loop on boundaries
	for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
	for(Bj=Bi, Bj++     ; Bj!=r_str.end(); Bj++) {
		//--- a) create unused_bnd (vertex list of unused boundaries)
		unused_bnd.clear();
		for(Bk=r_str.begin(); Bk!=r_str.end(); Bk++) {
			if(Bk!=Bi && Bk!=Bj) {
				unused_bnd.insert(unused_bnd.end(), Bk->r_vtx.begin(), Bk->r_vtx.end());
			}
		}
		
		//--- b) double loop on Bi and Bj children
		for(Ci=Bi->children.begin(); Ci!=Bi->children.end(); Ci++) {
		for(Cj=Bj->children.begin(); Cj!=Bj->children.end(); Cj++) {
			//----------rename Child_vtx_1 in Child_vtx_2 if necessary ------
			// necessary only if both Ci and Cj use the letter Child_vtx_1
			// this letter can appear only at the front of Ci and Cj boundary bribes
			// (case 1 in Boundary::compute_twoBoundaries_children)
			Line L_tmp=*Cj;
			if(!(Cj->empty()) && !(Ci->empty())) {
				if(Cj->front()==global::Child_vtx_1 && Ci->front()==global::Child_vtx_1) {
					//WARNING ! it is more reasonable to precise explicitly Vertex type in the replace function as it is necessary in Boundary::distinguish_low_up() and in Boundary::lower_to_upper()
					replace(L_tmp.begin(), L_tmp.end(), (Vertex) global::Child_vtx_1, (Vertex) global::Child_vtx_2);
				}
			}
			
			//-----union of two boundary's bribes------------------
			child.clear();
			
			child.insert(child.end(), Ci->begin(), Ci->end());      //add first bribe (Ci)
			child.push_back(global::Child_vtx_3);                   //add a new vertex
			child.insert(child.end(), L_tmp.begin(), L_tmp.end());  //add second bribe (Cj)
			child.push_back(global::Child_vtx_3);                   //add same new vertex
			child.push_back(global::End_Boundary);                  //add 'end of boundary'
			
			//add unused boundaries
			child.insert(child.end(), unused_bnd.begin(), unused_bnd.end());
			
			//add 'end of domain'
			child.push_back(global::End_Domain);
			
			//add the domains' child to the storage
			children.insert(child);
		}}
	}}
}

//create set of all possible partitions in two blocks
void Domain::partition(const list<Boundary>& l_bnd) {
  //note 1 : result is stored in class member : possible_partition
  //note 2 : we use two different algorithms to treat more efficiently 0. boundaries
	
	possible_partition.clear();
	
	Line r1;	//one possible partition's first block
	Line r2;	//one possible partition's second block
	
	//convention (as in math) : if boundaries list is empty, the only possible partition is (empty, empty)
	if (l_bnd.empty()) {
		possible_partition.insert(couple(r1,r2));
		return;
	}
	
	list<Boundary> l_normal_bnd;   //list of normal boundaries (all but 0.)
	
	set<couple> zero_partition;    //all possible two-blocks partitions of 0. boundaries
	set<couple> normal_partition;  //all possible two-blocks partitions of normal boundaries
	
	unsigned long n, i, j, q;
	list<Boundary>::const_iterator Bj;
	
	//-----step 1 : count and remove (0.) boundaries-------------------------------
	n=0;
	for(Bj=l_bnd.begin(); Bj!=l_bnd.end(); Bj++) {
		if(Bj->isSingleSpot()) {     //test if boundary == 0.
			n++;
		} else {
			l_normal_bnd.push_back(*Bj);
		}
	}
	
	//-----step 2 : partitions of 0. boundaries----------------------------
	if(n!=0) { //there is n+1 possible partitions
		for(i=0; i<n; i++) {
			r2.push_back(0);
			r2.push_back(global::End_Boundary);
		}
		zero_partition.insert( couple(r1,r2) );
		for(i=0; i<n; i++) {
			r1.push_back(0);
			r1.push_back(global::End_Boundary);
			r2.pop_back();
			r2.pop_back();
			//add this partition to the list of 0. boundaries partitions
			zero_partition.insert( couple(r1,r2) );
		}
	}
	
	//-----step 3 : partitions of normal boundaries------------------------
	n=l_normal_bnd.size();
	
	if (n!=0) {
		q=powers_of_two(n, true); //true means "handle overflow"
		
		//there is q possible partitions
		for(i=0;i<q;i++) {
			//bits of i "represents" the partition : 0 means first block and 1 means second one
			r1.clear();
			r2.clear();
			
			for(j=0, Bj=l_normal_bnd.begin(); j<n; j++, Bj++) {
				//we decide here whether j-th boundary is in the first or the second block
				//caution, you need to think 1 or 2 seconds to understand the following test
				if( ((i) & (powers_of_two(j, true))) == powers_of_two(j, true)) {
					r1.insert(r1.end(), Bj->r_vtx.begin(), Bj->r_vtx.end());
				} else {
					r2.insert(r2.end(), Bj->r_vtx.begin(), Bj->r_vtx.end());
				}
			}
			
			//add partition to the list of normal boundaries partition
			normal_partition.insert( couple(r1,r2) );
		}
	}
	
	//-----step 4 : complete set of partitions-----------------------------
	
	//if there is no "0. boundaries" partition
	if(zero_partition.empty()) {
		possible_partition=normal_partition;
		return;
	}
	
	//if there is no normal boundaries partition
	if(normal_partition.empty()) {
		possible_partition=zero_partition;
		return;
	}
	
	//both normal and zero boundaries exist
	set<couple>::iterator Ci;  //iterator on the set of partitions of zero boundaries
	set<couple>::iterator Cj;  //iterator on the set of partitions of normal boundaries
	
	//we have one complete partition for each couple of partitions of the zero and normal boundaries
	for(Ci=zero_partition.begin(); Ci!=zero_partition.end(); Ci++) {
	for(Cj=normal_partition.begin(); Cj!=normal_partition.end(); Cj++) {
		r1.clear();
		r2.clear();
		
		//r1 = (Ci->first) U (Cj->first)
		r1.insert(r1.end(), Ci->first.begin(), Ci->first.end());
		r1.insert(r1.end(), Cj->first.begin(), Cj->first.end());
		
		//r2 = (Ci->second) U (Cj->second)
		r2.insert(r2.end(), Ci->second.begin(), Ci->second.end());
		r2.insert(r2.end(), Cj->second.begin(), Cj->second.end());
		
		possible_partition.insert( couple(r1,r2) );
	}}
}


//-------------------------------------------------------------------------------------------------
// Children number estimation
//-------------------------------------------------------------------------------------------------

//estimates the number of children (by estimating the number of total possible moves)
unsigned int Domain::estimate_children_number() {
	update_representation();
	
	unsigned int estimation=0;		//estimation value
	
	list<Boundary>::const_iterator Bi, Bj;
	unsigned int vtx_number;		//number of vertex of a boundary
	
	//---------- number of moves linking a boundary to itself ------------------
	//for each boundary, the number of possible move is (number of vertex)^2*(number of partitions)
	//where (number of partitions) is the number of ways to distribute the other boundaries into two sets
	
	//note : we don't need to handle overflow when computing powers of two
	if (!r_str.empty()) {
		for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
			vtx_number = Bi->r_vtx.size()-1;
			estimation += vtx_number*vtx_number*powers_of_two(r_str.size()-1, false);
		}
	}
	
	//----------- number of moves linking two different boundaries --------------
	//for all couple (Bi, Bj) of boundaries,
	//there is (number of vertex of Bi)*(number of vertex of Bj) possible moves
	for(Bi=r_str.begin(); Bi!=r_str.end(); Bi++) {
	for(Bj=Bi, Bj++;      Bj!=r_str.end(); Bj++) {
		estimation += (Bi->r_vtx.size()-1)*(Bj->r_vtx.size()-1);
	}}
	
	return estimation;
}

//-------------------------------------------------------------------------
//lower-upper case distinction
//--------------------------------------------------------------------------

//Rename upper-case letters in lower-case ones (when needed)
void Domain::distinguish_low_up() {
	update_representation();
	change_structure();
	
	list<Boundary>::iterator Ci;
	for(Ci=r_str.begin(); Ci!=r_str.end(); Ci++) {
		Ci->distinguish_low_up();
	}
	
	update_representation();
}

//rename small letters in order of appearance
void Domain::rename_low_in_order() {
	update_representation();
	change_structure();
	
	list<Boundary>::iterator Ci;
	for(Ci=r_str.begin(); Ci!=r_str.end(); Ci++) {
		Ci->rename_low_in_order();
	}
	
	update_representation();
}

//rename lower-case letters into upper-case ones
//caution : structured representation must be uptodate
void Domain::lower_to_upper() {
	Vertex sav=Representation::free_upper_vtx;
	list<Boundary>::iterator Ci;
	for(Ci=r_str.begin(); Ci!=r_str.end(); Ci++) {
		Ci->lower_to_upper();
	}
	
	//set back the first free upper-case vertex
	// note : it will be used again during computation of the next domain (in Land class)
	// this is valid because a move takes part in only one domain
	Representation::free_upper_vtx=sav;
}
