/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "game.h"

//--------------------------------------------------------------------------
//Initialization
//--------------------------------------------------------------------------

//Initialization of default values
Game::Game() {
	//for analysis
	lives_min_analysis=false; lives_max_analysis=false;
	lives_min=0; lives_max=0;
	nimber_min_analysis=false; nimber_max_analysis=false;
	nimber_min=0; nimber_max=0;
}

//This function is run as a Qt thread when calling start()
void Game::run() {
	win_loss a=0; //computation result
	
	if(parameter.computation_kind==global::Analysis) {
		analysis_selection();
	} else{
		a=compute();
	}
	
	//computation is finished. Result is saved in computation_result
	game_lock.lock();
		computation_result=a;
	game_lock.unlock();
}

//Compute win/loss of a position
//(it can be a normal or a check computation)
win_loss Game::compute() {
  //initialize studied position
	Pos_nim pos_ini;
	
	if( parameter.use_given_position ){	//any chosen position
		pos_ini.first=parameter.given_position;
		pos_ini.second=parameter.given_nimber;
	}
	else{				//classic position with "0." spots
		Line l_ini;
		int i;
		for (i=0; i<parameter.spots_number; i++) {
			l_ini.push_back(0);
			l_ini.push_back(global::End_Boundary);
		}
		l_ini.push_back(global::End_Domain);
		l_ini.push_back(global::End_Land);
		l_ini.push_back(global::End_Position);
		pos_ini.first=l_ini;
		pos_ini.second=0;	//nimber=0
	}
	
  //initialize trace and graph objects
	trace_computation.initialize();
	check_graph.initialize();
	
  //starts computation
	if(parameter.computation_kind==global::Main){		//normal computation
		return compute_win_loss(pos_ini);
	}
	
	if(parameter.computation_kind==global::Check){		//check computation
		return check_win_loss(pos_ini);
	}
	
	return complete_check();				//complete_check
}


//--------------------------------------------------------------------------
//Functions used to choose the branch to explorate first
//They are used by Game::give_priority, which in turn is used in Game::compute_win_loss.
//--------------------------------------------------------------------------

//PositionInfo equal operator
bool PositionInfo::operator==(const PositionInfo& a) const{
	if (nb_nimlives != a.nb_nimlives)	return false;
	if (nb_lands != a.nb_lands)		return false;
	if (nb_moves != a.nb_moves)		return false;
	
	return (p_n.first != a.p_n.first);
}

//PositionInfo compare operator
bool PositionInfo::operator<(const PositionInfo& a) const{
	//priority to positions with minimal (nimber + number of lives)
	if (nb_nimlives != a.nb_nimlives) {
		return (nb_nimlives<a.nb_nimlives);
	}
	
	//priority to positions with a lot of lands
	if (nb_lands != a.nb_lands) {
		return (a.nb_lands<nb_lands);
	}
	
	//priority to positions with a little estimation of children number
	if(nb_moves != a.nb_moves) {
		return (nb_moves<a.nb_moves);
	}
	
	//lexicographical order
	return (p_n.first<a.p_n.first);
}

//Sort children (list "children") with priorities
//we modify this list given in parameter
void Game::give_priority(list<Pos_nim> &children) {
	//---------- compute information about children ------------------
	//create a list of children with their computed extra information
	list<PositionInfo> children_with_info;
	PositionInfo Pinfo_tmp;
	list<Pos_nim>::const_iterator Pi;
	
	Position Ptmp;
	for(Pi=children.begin(); Pi!=children.end(); Pi++) {
		Pinfo_tmp.p_n=*Pi;
		Pinfo_tmp.nb_nimlives=(Pi->second + nb_lives(Pi->first)); //nb_nimlives = nimber + number of lives
		Pinfo_tmp.nb_lands=count(Pi->first.begin(), Pi->first.end(), global::End_Land);
		
		Ptmp.reset(Pi->first);
		Pinfo_tmp.nb_moves=Ptmp.estimate_children_number();
		
		//add new information to the list
		children_with_info.push_back(Pinfo_tmp);
	}
	
	//---------- sort ------------------------------------------------
	children_with_info.sort();
	
	//---------- create result list ----------------------------------
	list<PositionInfo>::const_iterator Pj;
	children.clear();
	for(Pj=children_with_info.begin(); Pj!=children_with_info.end(); Pj++) {
		//add the child representation to the list of children
		children.push_back(Pj->p_n);
	}
}


//--------------------------------------------------------------------------
// Tiny and useful functions
//--------------------------------------------------------------------------

//canonization of a position described by a vertices list
//used only in check computation
void Game::cano(Line& r_vtxA) const {
	Position Ptmp(r_vtxA);
	Ptmp.cano();
	r_vtxA=Ptmp.r_vtx;
}

//returns the nimber resulting of a list of nimber, using "exclusive or"
//a complete game nimber is the xor of the nimber of its independant subgames (lands)
nimber Game::nimber_xor(list<nimber>& l_nimber) {
	nimber result=0;	//identity element for "exclusive or"
	list<nimber>::const_iterator Ni;
	
	for(Ni=l_nimber.begin(); Ni!=l_nimber.end(); Ni++) {
		result ^= *Ni;	//exclusive or (c++ operator)
		
		if(result>global::Nim_max){
			Error(QString("Too big nimber in Game::nimber_xor"), true);
		}
	}
	
	return result;
}

//Children of a Pos_nim :
// - if part_nim == false, returns children of position part
// - if part_nim == true, returns children of nimber part
// the children are stored in the 'result' parameter
void Game::pos_nim_children(const Pos_nim& pnA, bool part_nim, list<Pos_nim>& result) const {
	Line r_vtxA=pnA.first;
	nimber nimA=pnA.second;
	Pos_nim child; //child of pnA
	if(!part_nim) {
		//------- move in the position part -----------
		//normal children of r_vtxA
		Position Ptmp(r_vtxA);
		Ptmp.compute_children();
		
		set<Line>::iterator it;
		for(it=Ptmp.children.begin(); it!=Ptmp.children.end(); it++) {
			child.first=*it;
			child.second=nimA;	//the nimber part of each child is nimA
			result.push_back(child);
		}
	} else {
		//------- move in the nimber part -------------
		int i;
		for(i=0; i<nimA; i++) {
			child.first=r_vtxA;	//position part is unchanged
			child.second=i;
			result.push_back(child);
		}
	}
}

// - Returns "Loss" (nimber=0) or "Win" (nimber>=1) if position part is empty.
// - Returns "Nim_unknown" in other cases.
win_loss Game::wl_is_known(const Pos_nim& pnA) {
	if(pnA.first.size()==1) {
		//position part = empty position -> known nimber
		if(pnA.second==0){
			return global::Loss;
		}
		else{	//case nimber>=1
			return global::Win;
		}
	} else { //position part is not empty -> unknown nimber
		return global::Nim_unknown;
	}
}

//Add a position to the appropriate database
void Game::add_to_database(const Pos_nim &pn) {
	base_lock.lock();
		//add the position in the correct database (depends on computation kind)
		if(parameter.computation_kind==global::Main){
			main_base[toString(pn.first)]=pn.second;
		}else{
			check_base[toString(pn.first)]=pn.second;
		}
	base_lock.unlock();
}

//--------------------------------------------------------------------------
// Main functions
//--------------------------------------------------------------------------

//Simplify the given position with known nimbers
//caution : this function modifies the position-nimber given in parameter
void Game::apply_known_nimbers(Pos_nim& pnA, bool with_main_base) {
	Line r_vtxA=pnA.first;
	nimber nimA=pnA.second;
	
	//nothing to do if position part is empty
	if(Position(r_vtxA).position_type()==global::Empty_position) {
		return;
	}
	
	//list of lands (given by their vertex representation)
	//note : lands are listed as positions (list_lands add an End_Position character to each land)
	list<Line> lands=Position(r_vtxA).list_lands();
	
	// ------------- look for known lands in the database ----------------
	list<nimber> known_nimbers; //list of nimbers of known lands
	Line unknown_lands;  //positions's vertex representation keeping only the unknown lands
	
	map<string, nimber>::const_iterator it;
	bool found;
	list<Line>::iterator Li;
	
	//for each land, look for its nimber in the database
	for(Li=lands.begin(); Li!=lands.end(); Li++) {
		if(with_main_base){
			it = main_base.find(toString(*Li));
			found=(it!=main_base.end());
		}else {
			it = check_base.find(toString(*Li));
			found=(it!=check_base.end());
		}
		if(found) {
			//known land
			known_nimbers.push_back(it->second);
		} else {
			//unknown land
			unknown_lands.insert(unknown_lands.end(), Li->begin(), Li->end());
			unknown_lands.pop_back(); //delete position's end character (End_Position)
		}
	}
	
	unknown_lands.push_back(global::End_Position);
	pnA.first=unknown_lands;
	
	if(!known_nimbers.empty()) {
		known_nimbers.push_back(nimA);        //add nimber part to the list of nimbers
		pnA.second=nimber_xor(known_nimbers); //compute resulting nimber with xor
	} else {
		//known_nimbers is empty : nimber part remains unchanged
		pnA.second=nimA;
	}
}

//compute nimber for every land, except the biggest
//function always called on a position with >=2 land
// caution : this function modifies the position-nimber given in parameter
void Game::kill_little_lands(Pos_nim& pnA) {
	list<Line> list_of_land=Position(pnA.first).list_lands();
	list<Line>::iterator L_big=list_of_land.begin();
	list<Line>::iterator Li=list_of_land.begin(); Li++; //increment is safe because there is always at least two lands
	list<nimber> known_nimbers;
	nimber one_land_nimber;
	Line pos1;
	while(Li!=list_of_land.end()){
		if(nb_lives(*Li)>nb_lives(*L_big)){
			//current land is the new bigger one
			pos1=*L_big;
			L_big=Li;
		}else{
			pos1=*Li;
		}
		one_land_nimber=compute_nimber(pos1);
		if(one_land_nimber==global::Computation_end){	//in case of stop or zapping
			return;
		}
		known_nimbers.push_back(one_land_nimber);
		Li++;
	}
	known_nimbers.push_back(pnA.second);	//original nimber part
	pnA.second=nimber_xor(known_nimbers);	//new nimber computing
	pnA.first=*L_big;
}

//compute the nimber of a position : tries (posA;0), (posA;1), (posA;2)...
//until (pos;nim) is a loss to find the position's nimber ("0-1-2 algorithm")
// - returns the nimber of posA
// - returns "Computation_end" in case of interruption or zapping
nimber Game::compute_nimber(const Line& r_vtxA) {
	//stop computation if asked by user
	if (trace_computation.stop_signal()) return global::Computation_end;
	
	Pos_nim pn;
	pn.first=r_vtxA;
	pn.second=0;
	bool nimber_found=false;
	win_loss a_try;
	int current_pos_number=0;
	
	Level_trace one_level(&trace_computation, true); //object one_level is destroyed at the end of function

	trace_computation.add(pn);

	//note : this loop ends when user emits signal (because compute_win_loss returns Computation_end)
	while(!nimber_found){
		a_try=compute_win_loss(pn);
		
		//return immediatly to above level on user demand
		if (trace_computation.user_interrupt(true)) {
			return global::Computation_end;
		}
		
		if(a_try==global::Loss){
			nimber_found=true;
		}
		else if(a_try!=global::Computation_end) {	//a_try==Computation_end if zapping on this level
			pn.second++;
			if(pn.second>global::Nim_max){
				Error(QString("Too big nimber in Game::compute_nimber"), true);
			}
			current_pos_number++;
			trace_computation.add(pn);
			trace_computation.change_studied_pos(current_pos_number);
		}
	}
	
	return pn.second;
}

//Compute win/loss of a position (recursive function)
// - returns "Win" if at least one child is a "losing" one
// - returns "Loss" if all children are "winning" ones
// - returns "Computation_end" in case of interruption or zapping
win_loss Game::compute_win_loss(const Pos_nim& pnA) {
	//stop computation if asked by user
	if (trace_computation.stop_signal()) return global::Computation_end;
	
	//-----step 1 : find nimber for already known and little lands-----------------------------
	Pos_nim simp=pnA;	//simp means "simplified" position
	
	apply_known_nimbers(simp,true);	//replace already known lands by their nimber
	
	if(wl_is_known(simp) != global::Nim_unknown){	//nimber is known for every land
		return wl_is_known(simp);
	}
	
	if(Position(simp.first).position_type()==global::Multi_lands){
		kill_little_lands(simp); //compute nimber for every land except the biggest
	}
	
	//user-signal reception after kill_little_lands call
	if (trace_computation.user_interrupt(false)) {
		return global::Computation_end;
	}
	
	//-----step 2 : compute children-----------------------------------------------------------
	list<Pos_nim> children;
	list<Pos_nim> children_pos;
	
	//pos_nim_children parameters : position, nimber/position part, result variable
	pos_nim_children(simp, true, children);       //children of the nimber part
	pos_nim_children(simp, false, children_pos);  //children of the position part
	
	//canonization of children's position part
	//separating children computation and canonization allows to save time, because certain children won't be
	//canonized if a child is found to be a losing one
	list<Pos_nim>::iterator Pi=children_pos.begin();
	Position Ptmp;
	while(Pi!=children_pos.end()) {
		Ptmp.reset(Pi->first);
		Ptmp.cano();
		Pi->first=Ptmp.r_vtx;
		
		apply_known_nimbers(*Pi,true);	//replace already known land by their nimber
		if(wl_is_known(*Pi)==global::Loss) return global::Win; //returns "Win" if any child is "Loss"
		if(wl_is_known(*Pi)==global::Win) {	//Winning children are useless
			Pi=children_pos.erase(Pi);
		}
		else {
			Pi++;
		}
	}
	
	//sort children of the position part
	give_priority(children_pos);	//caution : children_pos is changed
	children_pos.unique();		//removes all consecutive duplicate positions
	
	//union of all children (nimber part + position part)
	children.insert(children.end(),children_pos.begin(),children_pos.end());
	
	//-----step 3 (main step): recursive computation on children-------------------------------
	//-----compute recursively the state of each child until one child is a losing one---------
	
	//create a new level to trace computing (automatically destroyed when function returns)
	Level_trace one_level(&trace_computation, false);
	
	//initialize trace object with the children
	trace_computation.add(children);
	
	int children_number=children.size();
	int unknown_number=children_number;	//number of still alive children
	
	//vector of booleans to remember children status (known/unknown)
	vector<bool> known_status(children_number);
	fill(known_status.begin(), known_status.end(), false);
	vector<bool>::iterator Pi_known_status;
	
	//the three following parameters are linked
	Pi=children.begin();
	Pi_known_status=known_status.begin();
	int studied_pos_index=0;
	
	bool take_one_step_forward=false;		//go to next position (even if current position is unknown)
	bool take_one_step_backward=false;	//go to previous position
	int wanted_child_index=-1;
	
	while(unknown_number>0) {
		//compute another child (as part of normal computation or as asked by user)
		if(take_one_step_backward){	//case of a backward zapping
			while(*Pi_known_status || take_one_step_backward) {
				take_one_step_backward=false;
				if(studied_pos_index==0){
					studied_pos_index=children_number-1;
					Pi=children.end(); Pi--;
					Pi_known_status=known_status.end(); Pi_known_status--;
				} else {
					studied_pos_index--; Pi--; Pi_known_status--;
				}
			}
		} else {		//case of no zapping, of a normal zapping, or a zapping from children's table
			while(*Pi_known_status || take_one_step_forward || wanted_child_index!=-1) {
				take_one_step_forward=false;
				if(studied_pos_index==children_number-1){
					studied_pos_index=0; Pi=children.begin(); Pi_known_status=known_status.begin();
				} else {
					studied_pos_index++; Pi++; Pi_known_status++;
				}
				if(studied_pos_index==wanted_child_index || wanted_child_index>=children_number){
					wanted_child_index=-1; //child wanted found || click too low in the children's table (should not happen)
				}
			}
		}
		trace_computation.change_studied_pos(studied_pos_index);
		
		win_loss result=compute_win_loss(*Pi);
		
		//user-signal reception
		int val=trace_computation.user_request();
		if(val==-1) {		//zapping above or computation stopped
			return global::Computation_end;
		} else if (val==-2) {	//simple zapping at this level
			take_one_step_forward=true;
		} else if (val==-3) {	//zapping backward
			take_one_step_backward=true;
		} else if (val!=-4) {	//zapping from children's table
			wanted_child_index=val;
		}
		
		//-----case of a losing child -----
		if(result==global::Loss)  {
			//end of function. One losing child means parent is a winning position
			return global::Win;
		}
		
		//-----case of a winning child -----
		if(result!=global::Computation_end){
			unknown_number--;	//one child was killed (computed winning)
			trace_computation.known_child(studied_pos_index); //change child state in trace object (->status known)
			*Pi_known_status=true;		//change child state in local object (->status known)
		}
	}
	
	//if we get there, it means all children were winning ones, so "simp" is a losing position
	//the position part's nimber is then equal to the nimber part - we add this result in database
	add_to_database(simp);
	
	return global::Loss;
}
