/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GAME_H
#define GAME_H

#include <qthread.h>
#include <qmutex.h>

#include "position.h"
#include "graph.h"
#include "trace.h"

//little class used when sorting children
//it contains a position with extra information
class PositionInfo {
  public:
	Pos_nim p_n;
	int nb_lands;
	int nb_moves;
	int nb_nimlives;	//nb_nimlives = nimber + number of lives
	
	bool operator==(const PositionInfo& a) const;
	bool operator<(const PositionInfo& a) const;
};

class Parameter {
  public:	
	int spots_number;           //number of starting spots
	bool use_given_position;    //when true, user can define the starting position
	Line given_position;        //user-defined starting position (position part)
	nimber given_nimber;        //user-defined starting extended position (nimber part)
	
	int computation_kind;       //kind of computation : 0 = normal, 1 = check, 2 = analysis
};


//Program's main class
class Game : public QThread {
	
  public:
	Game();
	
	//----------databases----------------------------------------------------------------------
	map<string, nimber> main_base;      //for normal computation results
	map<string, nimber> check_base;     //for check computation results
	map<string, nimber> analysis_base;  //for an analysis of positions in main_base
	
	//----------tracing------------------------------------------------------------------------
	Trace trace_computation;
	
	//----------communication------------------------------------------------------------------
	nimber computation_result;          //result of a computation
	bool check_failed;                  //indicate that a check computation failed
	string check_failure_msg;           //message explaining why check computation failed
	
	//----------QThreads-----------------------------------------------------------------------
	virtual void run();                 //this function is called as a Qt thread by start()
	QMutex base_lock, game_lock;        //multi-thread locks to secure access to various variables
	
	//----------computation parameters---------------------------------------------------------
	Parameter parameter;
	
	//----------launching function-------------------------------------------------------------
	win_loss compute();                   //launch computation using the given parameters
	
	//----------tiny and useful functions------------------------------------------------------
	void cano(Line& r_vtxA) const;			//canonization of a position given by its vertices list
	nimber nimber_xor(list<nimber>& l_nimber);	//nimber resulting from a list of nimber
	void pos_nim_children(const Pos_nim& posA, bool part_nim, list<Pos_nim> & result) const;
	win_loss wl_is_known(const Pos_nim& posA);
	void add_to_database(const Pos_nim&);
	
	//----------normal computation-------------------------------------------------------------
	void give_priority(list<Pos_nim> &s);				//sort children
	void apply_known_nimbers(Pos_nim& posA, bool with_main_base);	//use known nimbers to simplify a Pos_nim
	void kill_little_lands(Pos_nim& pnA);				//compute nimber of every land but the biggest
	nimber compute_nimber(const Line& r_vtxA);			//compute the nimber of posA using 0-1-2 algorithm
	win_loss compute_win_loss(const Pos_nim& posA);			//recursively compute Win/Loss of posA
	
	//----------check computation--------------------------------------------------------------
	win_loss check_win_loss(Pos_nim pnA);
	bool kill_known_lands(Pos_nim& pnA);
	bool check_expected_win_loss(Pos_nim pnA, win_loss expected_winloss);
	void check_failure(Pos_nim pn, int failure_kind);
	win_loss complete_check();
	
	Graph check_graph;	//complete graph of a check computation
	
	//----------analysis-----------------------------------------------------------------------
	bool lives_min_analysis, lives_max_analysis, nimber_min_analysis, nimber_max_analysis;	//active/inactive parameters
	int lives_min, lives_max, nimber_min, nimber_max;					//parameters value
	void analysis_selection();	//put in analysis database the positions of "main_base" which satisfy user choice
};

#endif
