/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "glopwidget.h"
#include "widgetpopup.h"

#include <QtGui>
#include <QString>

//--------------------------------------------------------------------------
//Interface definition
//--------------------------------------------------------------------------
GlopWidget::GlopWidget( QWidget *parent) : QMainWindow( parent ) {
	//window title
	setWindowTitle(QString("GLOP"));
	//game object
	game = new Game;
	
	//base is empty ("saved state")
	mainbaseisSaved = true;
	checkbaseisSaved = true;
	computeRunning = false;
	
	//timer
	impulse = new QTimer(this);
	connect( this, SIGNAL(isComputing(bool)), this, SLOT(activate_timer(bool)) );
	
	//on each impulse, check result state and refresh tables tracing computation
	connect( impulse, SIGNAL(timeout()), this, SLOT(wait_result()) );
	connect( impulse, SIGNAL(timeout()), this, SLOT(trace_refresh()) );
	
	//error handling (check error on each timer impulse)
	WidgetError *widget_error= new WidgetError(this);
	connect( impulse, SIGNAL(timeout()), widget_error, SLOT(receive_error()) );
	
	//connexion mapping
	fileSaveAs_Mapper=new QSignalMapper(this);
	startCompute_Mapper=new QSignalMapper(this);
	purgeBase_Mapper=new QSignalMapper(this);
	readText_Mapper=new QSignalMapper(this);
	connect( fileSaveAs_Mapper, SIGNAL(mapped(int)), this, SLOT(fileSaveAs(int)) );
	connect( startCompute_Mapper, SIGNAL(mapped(int)), this, SLOT(start_computation(int)) );
	connect( purgeBase_Mapper, SIGNAL(mapped(int)), this, SLOT(purgeBase(int)) );
	connect( readText_Mapper, SIGNAL(mapped(int)), this, SLOT(read_text_file(int)) );
	
	//layout
	QWidget *widget_central = new QWidget( this );
	setCentralWidget(widget_central);
	QVBoxLayout *central_Vlayout = new QVBoxLayout(widget_central);	//vertical layout of main window
	
	//-----upper box-------------------------
	QHBoxLayout *upper_box = new QHBoxLayout;	//horizontal layout of top part
	central_Vlayout->addLayout( upper_box );
	
	//sprouts icon
	QLabel *labelIcon = new QLabel( widget_central );
	QPixmap *pixIcon = new QPixmap(":/images/sprouts32.png");
	labelIcon->setPixmap(*pixIcon);
	labelIcon->setFixedSize( 32, 32 );
	upper_box->addWidget( labelIcon );
	
	//main, check, graph and analysis database size
	mainSizeLabel = new QLabel(widget_central);
	checkSizeLabel = new QLabel(widget_central);
	graphSizeLabel = new QLabel(widget_central);
	analysisSizeLabel = new QLabel(widget_central);
	
	upper_box->addWidget(mainSizeLabel);
	upper_box->addWidget(checkSizeLabel);
	upper_box->addWidget(graphSizeLabel);
	upper_box->addWidget(analysisSizeLabel);
	
	//refresh size (to make the widget appear)
	refresh_base_size(global::Main);
	refresh_base_size(global::Check);
	refresh_base_size(global::Graph);
	refresh_base_size(global::Analysis);
	
	//Stop button
	QPushButton *stop_button = new QPushButton( "Stop", widget_central );
	stop_button->setEnabled(false);
	stop_button->setFixedSize(stop_button->sizeHint());
	upper_box->addWidget( stop_button);
	connect( stop_button, SIGNAL(clicked()), this, SLOT(stop_computing()) );
	connect( this, SIGNAL(isComputing(bool)), stop_button, SLOT(setEnabled(bool)) );
	
	//-----Tabs in the main window-----------
	tabWidget = new QTabWidget;
	central_Vlayout->addWidget( tabWidget );
	setup_parameters_tab();
	setup_compute_tab();
	setup_test_tab();
	setup_analysis_tab();
	setup_info_tab();
}

//Parameters tab
void GlopWidget::setup_parameters_tab() {
  //Main widget, layout and connection of this tab
	QWidget *parameters_tab = new QWidget;
	//Tab disabled when computing
	connect( this, SIGNAL(isComputing(bool)), parameters_tab, SLOT(setDisabled(bool)) );
	//Vertical Layout
	QVBoxLayout *parameters_tab_Vlayout = new QVBoxLayout( parameters_tab );
	
  //Choose Position to compute
	QGroupBox *choosePositionBox = new QGroupBox("Position to compute", parameters_tab);
	parameters_tab_Vlayout->addWidget( choosePositionBox );
	QGridLayout *grid_ptc = new QGridLayout( choosePositionBox );
	//classic start position
	classic_position = new QRadioButton("Classic start position", choosePositionBox);
	grid_ptc->addWidget( classic_position, 0, 0 );
	QLabel *spots_number_title = new QLabel( choosePositionBox );
	spots_number_title->setText("Number of spots :");
	grid_ptc->addWidget( spots_number_title, 0, 1 );
	box_spots_number = new QSpinBox( choosePositionBox );
	box_spots_number->setMinimum(2);
	box_spots_number->setMaximum(99);
	box_spots_number->setFixedSize(box_spots_number->sizeHint());
	box_spots_number->setValue(11);
	grid_ptc->addWidget( box_spots_number, 0, 2 );
	//free start position
	given_position = new QRadioButton("Free start position", choosePositionBox);
	grid_ptc->addWidget( given_position, 1, 0 );
	QLabel *repr = new QLabel( choosePositionBox );
	repr->setText("Vertex representation :");
	grid_ptc->addWidget( repr, 1, 1 );
	box_given_position = new QLineEdit( choosePositionBox );
	grid_ptc->addWidget( box_given_position, 1, 2 );
	box_given_position->setText( QString("0.0.0.}]0.A.}0.A.}]!") );
	box_given_position->setMinimumWidth( 150 );
	QLabel *nimb_lab = new QLabel( choosePositionBox );
	nimb_lab->setText("Nimber :");
	grid_ptc->addWidget( nimb_lab, 2, 1 );
	box_given_nimber = new QSpinBox( choosePositionBox );
	box_given_nimber->setMinimum(0);
	box_given_nimber->setMaximum(99);
	box_given_nimber->setFixedSize(box_given_nimber->sizeHint());
	grid_ptc->addWidget( box_given_nimber, 2, 2 );
	//default case : classic start position
	classic_position->setChecked(TRUE);
	
  //Options
	QGroupBox *optionsBox = new QGroupBox("Options", parameters_tab);
	parameters_tab_Vlayout->addWidget( optionsBox );
	QGridLayout *grid_opt = new QGridLayout( optionsBox );
	//refresh time
	QLabel *refresh_lab = new QLabel( optionsBox );
	refresh_lab->setText("Refresh time (x 100 ms) :");
	grid_opt->addWidget( refresh_lab, 0, 0 );
	box_refresh = new QSpinBox( optionsBox );
	box_refresh->setMinimum(1);
	box_refresh->setMaximum(10);
	box_refresh->setValue(1);
	box_refresh->setFixedSize(box_refresh->sizeHint());
	grid_opt->addWidget( box_refresh, 0, 1 );
	//use graph
	QLabel *use_graph_lab = new QLabel( optionsBox );
	use_graph_lab->setText("Use graph :");
	grid_opt->addWidget( use_graph_lab, 1, 0);
	box_use_graph = new QCheckBox( optionsBox );
	grid_opt->addWidget( box_use_graph, 1, 1);
	
  //Graph options
	QGroupBox *graph_options = new QGroupBox("Graph options", parameters_tab);
	parameters_tab_Vlayout->addWidget( graph_options );
	QGridLayout *grid_graph_opt = new QGridLayout( graph_options );
	//use colors
	QLabel *use_colors_lab = new QLabel( graph_options );
	use_colors_lab->setText("Use colors :");
	grid_graph_opt->addWidget( use_colors_lab, 0, 0);
	box_use_colors = new QCheckBox( graph_options );
	grid_graph_opt->addWidget( box_use_colors, 0, 1);
	//detail level
	QLabel *detail_label = new QLabel( graph_options );
	detail_label->setText("Detail level :");
	grid_graph_opt->addWidget( detail_label, 1, 0 );
	box_detail = new QSpinBox( optionsBox );
	box_detail->setMinimum(1);
	box_detail->setMaximum(3);
	box_detail->setValue(2);
	box_detail->setFixedSize(box_detail->sizeHint());
	grid_graph_opt->addWidget( box_detail, 1, 1 );
	//minimal number of lives
	QLabel *mlg_lab = new QLabel( graph_options );
	mlg_lab->setText("Minimal number of lives :");
	grid_graph_opt->addWidget( mlg_lab, 2, 0 );
	box_mlg = new QSpinBox( optionsBox );
	box_mlg->setMinimum(0);
	box_mlg->setMaximum(99);
	box_mlg->setValue(0);
	box_mlg->setFixedSize(box_mlg->sizeHint());
	grid_graph_opt->addWidget( box_mlg, 2, 1 );
	//status (default : disabled, activated with the above checkbox)
	graph_options->setEnabled(false);
	connect( box_use_graph, SIGNAL(toggled(bool)), graph_options, SLOT(setEnabled(bool)) );
	
  //add Parameter tab to the tabWidget
	tabWidget->addTab(parameters_tab, "Parameters");
}

//Computation tab
void GlopWidget::setup_compute_tab() {
  //Layout for this tab
	QWidget *compute_tab = new QWidget;			//main widget of Computation tab
	connect( this, SIGNAL(isChecking(bool)), compute_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *compute_tab_Vlayout = new QVBoxLayout( compute_tab );	//layout for this tab
	compute_tab_Vlayout->setSpacing(8);
	compute_tab_Vlayout->setMargin(5);
	
  //Buttons
	QWidget *buttons_widget = new QWidget( compute_tab);
	compute_tab_Vlayout->addWidget( buttons_widget );
	QGridLayout *grid_buttons = new QGridLayout( buttons_widget );
	grid_buttons->setMargin(0);
	connect( this, SIGNAL(isComputing(bool)), buttons_widget, SLOT(setDisabled(bool)) );
	
	QPushButton *compute_button = new QPushButton( "Compute", compute_tab );
	compute_button->setFixedSize(compute_button->sizeHint());
	grid_buttons->addWidget( compute_button, 0, 0 );
	connect( compute_button, SIGNAL(clicked()), startCompute_Mapper, SLOT(map()) );
	startCompute_Mapper->setMapping(compute_button, global::Main);
	
	QPushButton *add_base = new QPushButton( "Add a database", compute_tab );
	add_base->setFixedSize(add_base->sizeHint());
	grid_buttons->addWidget( add_base, 1, 0 );
	connect( add_base, SIGNAL(clicked()), this, SLOT(fileOpen()) );
	
	QPushButton *save_base = new QPushButton( "Save main database", compute_tab );
	save_base->setFixedSize(save_base->sizeHint());
	grid_buttons->addWidget( save_base, 2, 0 );
	connect( save_base, SIGNAL(clicked()), fileSaveAs_Mapper, SLOT(map()) );
	fileSaveAs_Mapper->setMapping(save_base, global::Main);
	
	QPushButton *check_button = new QPushButton( "Check", compute_tab );
	check_button->setFixedSize(check_button->sizeHint());
	grid_buttons->addWidget( check_button, 0, 1 );
	connect( check_button, SIGNAL(clicked()), startCompute_Mapper, SLOT(map()) );
	startCompute_Mapper->setMapping(check_button, global::Check);
	
	QPushButton *save_check_base = new QPushButton( "Save check-database", compute_tab );
	save_check_base->setFixedSize(save_check_base->sizeHint());
	grid_buttons->addWidget( save_check_base, 1, 1 );
	connect( save_check_base, SIGNAL(clicked()), fileSaveAs_Mapper, SLOT(map()) );
	fileSaveAs_Mapper->setMapping(save_check_base, global::Check);
	
	QPushButton *save_check_graph = new QPushButton( "Save check graph", compute_tab );
	save_check_graph->setFixedSize(save_check_graph->sizeHint());
	grid_buttons->addWidget( save_check_graph, 2, 1 );
	connect( save_check_graph, SIGNAL(clicked()), this, SLOT(fileSaveGraph()) );
	save_check_graph->setEnabled(false);
	connect( box_use_graph, SIGNAL(toggled(bool)), save_check_graph, SLOT(setEnabled(bool)) );
	
	QPushButton *complete_check_button = new QPushButton( "Complete Check", compute_tab );
	complete_check_button->setFixedSize(complete_check_button->sizeHint());
	grid_buttons->addWidget( complete_check_button, 0, 2 );
	connect( complete_check_button, SIGNAL(clicked()), startCompute_Mapper, SLOT(map()) );
	startCompute_Mapper->setMapping(complete_check_button, global::Complete_check);
	
	QPushButton *purge_main_base = new QPushButton( "Purge main database", compute_tab );
	purge_main_base->setFixedSize(purge_main_base->sizeHint());
	grid_buttons->addWidget( purge_main_base, 1, 2 );
	connect( purge_main_base, SIGNAL(clicked()), purgeBase_Mapper, SLOT(map()) );
	purgeBase_Mapper->setMapping(purge_main_base, global::Main);
	
	QPushButton *purge_check_base = new QPushButton( "Purge check-database", compute_tab );
	purge_check_base->setFixedSize(purge_check_base->sizeHint());
	grid_buttons->addWidget( purge_check_base, 2, 2 );
	connect( purge_check_base, SIGNAL(clicked()), purgeBase_Mapper, SLOT(map()) );
	purgeBase_Mapper->setMapping(purge_check_base, global::Check);
	
  //Table to trace computation
	widget_trace=new WidgetTrace(compute_tab, &game->trace_computation);
	connect( this, SIGNAL(isComputing(bool)), widget_trace->table, SLOT(setEnabled(bool)) );
	compute_tab_Vlayout->addWidget( widget_trace );
	
  //add Computation tab to tabWidget
	tabWidget->addTab( compute_tab, "Computation" );
}

//Analysis tab
void GlopWidget::setup_analysis_tab() {
	QWidget *analysis_tab = new QWidget;
	connect( this, SIGNAL(isComputing(bool)), analysis_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *analysis_tab_Vlayout = new QVBoxLayout( analysis_tab );
	QLabel *filterTitle = new QLabel( analysis_tab );
	filterTitle->setText( "Filter database with the following parameters :" );
	filterTitle->setFixedSize(filterTitle->sizeHint());
	analysis_tab_Vlayout->addWidget( filterTitle );
	
  //Number of lives
	QGroupBox *livesGroup = new QGroupBox("Number of lives", analysis_tab);
	analysis_tab_Vlayout->addWidget( livesGroup );
	QGridLayout *lives_grid = new QGridLayout( livesGroup );
	
	//Minimum number of lives
	check_lives_min = new QCheckBox( livesGroup );
	lives_grid->addWidget( check_lives_min, 0, 0 );
	
	QLabel *lives_min_label = new QLabel( livesGroup );
	lives_min_label->setText("Minimum number of lives :");
	lives_min_label->setEnabled(false);
	connect( check_lives_min, SIGNAL(toggled(bool)), lives_min_label, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( lives_min_label, 0, 1 );
	
	box_lives_min = new QSpinBox( livesGroup );
	box_lives_min->setMinimum(0);
	box_lives_min->setMaximum(99);
	box_lives_min->setFixedSize(box_lives_min->sizeHint());
	box_lives_min->setEnabled(false);
	connect( check_lives_min, SIGNAL(toggled(bool)), box_lives_min, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( box_lives_min, 0, 2 );
	
	//Maximum number of lives
	check_lives_max = new QCheckBox( livesGroup );
	lives_grid->addWidget( check_lives_max, 1, 0 );
	
	QLabel *lives_max_label = new QLabel( livesGroup );
	lives_max_label->setText("Maximum number of lives :");
	lives_max_label->setEnabled(false);
	connect( check_lives_max, SIGNAL(toggled(bool)), lives_max_label, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( lives_max_label, 1, 1 );
	
	box_lives_max = new QSpinBox( livesGroup );
	box_lives_max->setMinimum(0);
	box_lives_max->setMaximum(99);
	box_lives_max->setFixedSize(box_lives_max->sizeHint());
	box_lives_max->setEnabled(false);
	connect( check_lives_max, SIGNAL(toggled(bool)), box_lives_max, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( box_lives_max, 1, 2 );
	
  //Nimber
	QGroupBox *nimberGroup = new QGroupBox("Nimber", analysis_tab);
	analysis_tab_Vlayout->addWidget( nimberGroup );
	QGridLayout *nimber_grid = new QGridLayout( nimberGroup );
	
	//Minimum nimber
	check_nimber_min = new QCheckBox( nimberGroup );
	nimber_grid->addWidget( check_nimber_min, 0, 0 );
	
	QLabel *label_nimber_min = new QLabel( nimberGroup );
	label_nimber_min->setText("Minimum nimber :");
	label_nimber_min->setEnabled(false);
	connect( check_nimber_min, SIGNAL(toggled(bool)), label_nimber_min, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( label_nimber_min, 0, 1 );
	
	box_nimber_min = new QSpinBox( nimberGroup );
	box_nimber_min->setMinimum(0);
	box_nimber_min->setMaximum(99);
	box_nimber_min->setFixedSize(box_nimber_min->sizeHint());
	box_nimber_min->setEnabled(false);
	connect( check_nimber_min, SIGNAL(toggled(bool)), box_nimber_min, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( box_nimber_min, 0, 2 );
	
	//Maximum nimber
	check_nimber_max = new QCheckBox( nimberGroup );
	nimber_grid->addWidget( check_nimber_max, 1, 0 );
	
	QLabel *label_nimber_max = new QLabel( nimberGroup );
	label_nimber_max->setText("Maximum nimber :");
	label_nimber_max->setEnabled(false);
	connect( check_nimber_max, SIGNAL(toggled(bool)), label_nimber_max, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( label_nimber_max, 1, 1 );
	
	box_nimber_max = new QSpinBox( nimberGroup );
	box_nimber_max->setMinimum(0);
	box_nimber_max->setMaximum(99);
	box_nimber_max->setFixedSize(box_nimber_max->sizeHint());
	box_nimber_max->setEnabled(false);
	connect( check_nimber_max, SIGNAL(toggled(bool)), box_nimber_max, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( box_nimber_max, 1, 2 );
	
  //Buttons
	QHBoxLayout *buttons_Hlayout = new QHBoxLayout;
	analysis_tab_Vlayout->addLayout( buttons_Hlayout );
	
	QPushButton *start_button = new QPushButton( "Start", analysis_tab);
	start_button->setFixedSize(start_button->sizeHint());
	connect( start_button, SIGNAL(clicked()), startCompute_Mapper, SLOT(map()) );
	startCompute_Mapper->setMapping(start_button, global::Analysis);
	buttons_Hlayout->addWidget( start_button );
	
	QPushButton *save_button = new QPushButton( "Save analysis database", analysis_tab);
	save_button->setFixedSize(save_button->sizeHint());
	buttons_Hlayout->addWidget( save_button );
	connect( save_button, SIGNAL(clicked()), fileSaveAs_Mapper, SLOT(map()) );
	fileSaveAs_Mapper->setMapping(save_button, global::Analysis);
	
	tabWidget->addTab( analysis_tab, "Analysis" );
}

//Test tab
void GlopWidget::setup_test_tab() {
	QWidget *test_tab = new QWidget;
	connect( this, SIGNAL(isComputing(bool)), test_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *test_tab_Vlayout = new QVBoxLayout( test_tab );
  //Children computation
	QGroupBox *children_group = new QGroupBox("Children computation", test_tab);
	test_tab_Vlayout->addWidget( children_group );
	QGridLayout *grid_test = new QGridLayout( children_group );
	
	QLabel *repr_2 = new QLabel( children_group );
	repr_2->setText("Position :");
	grid_test->addWidget( repr_2, 0, 0 );
	box_given_position_2 = new QLineEdit( children_group );
	grid_test->addWidget( box_given_position_2, 0, 1 );
	box_given_position_2->setText( QString("0.0.0.0.AB.}0.0.0.0.AB.}]!") );
	box_given_position_2->setMinimumWidth( 150 );
	
	QPushButton *children_button = new QPushButton( "Compute children", children_group );
	children_button->setFixedSize(children_button->sizeHint());
	grid_test->addWidget( children_button, 1, 0 );
	connect( children_button, SIGNAL(clicked()), this, SLOT(print_children()) );
	
	textEdit_children = new QTextEdit( children_group );
	textEdit_children->setReadOnly( TRUE );
	grid_test->addWidget( textEdit_children, 1, 1 );
	
	tabWidget->addTab( test_tab, "Children" );
}

//Info tab
void GlopWidget::setup_info_tab() {
	QWidget *info_tab = new QWidget;
	QGridLayout *grid_info = new QGridLayout( info_tab );
	
  //Images
	QLabel *labelHelp = new QLabel( info_tab );
	QPixmap *pixHelp = new QPixmap(":/images/help64.png");
	labelHelp->setPixmap(*pixHelp);
	labelHelp->setFixedSize( 64, 64 );
	grid_info->addWidget( labelHelp, 0, 0 );
	
	QLabel *labelLicence = new QLabel( info_tab );
	QPixmap *pixLicence = new QPixmap(":/images/licence64.png");
	labelLicence->setPixmap(*pixLicence);
	labelLicence->setFixedSize( 64, 64 );
	grid_info->addWidget( labelLicence, 0, 1 );
	
	QLabel *labelAbout = new QLabel( info_tab );
	QPixmap *pixAbout = new QPixmap(":/images/about64.png");
	labelAbout->setPixmap(*pixAbout);
	labelAbout->setFixedSize( 64, 64 );
	grid_info->addWidget( labelAbout, 0, 2 );
	
  //Buttons
	QPushButton *help_button = new QPushButton( "Help", info_tab );
	help_button->setFixedSize(help_button->sizeHint());
	grid_info->addWidget( help_button, 1, 0 );
	connect( help_button, SIGNAL(clicked()), readText_Mapper, SLOT( map() ) );
	readText_Mapper->setMapping(help_button, 0);
	
	QPushButton *licence_button = new QPushButton( "Licence", info_tab );
	licence_button->setFixedSize(licence_button->sizeHint());
	grid_info->addWidget( licence_button, 1, 1 );
	connect( licence_button, SIGNAL(clicked()), readText_Mapper, SLOT( map() ) );
	readText_Mapper->setMapping(licence_button, 1);
	
	QPushButton *about_button = new QPushButton( "About", info_tab );
	about_button->setFixedSize(about_button->sizeHint());
	grid_info->addWidget( about_button, 1, 2 );
	connect( about_button, SIGNAL(clicked()), readText_Mapper, SLOT( map() ) );
	readText_Mapper->setMapping(about_button, 2);
	
	tabWidget->addTab( info_tab, "Information" );
}


//--------------------------------------------------------------------------
//Starts/Stops computation
//--------------------------------------------------------------------------

//Starts a computation.
void GlopWidget::start_computation(int computation_kind) {
  //parameters initialization
	game->parameter.computation_kind=computation_kind;
	
	if (computation_kind==global::Check || computation_kind==global::Complete_check) {
		game->check_failed=false;
	}
	
	//Read user-controled Widget parameters
	if(computation_kind==global::Analysis) {
		read_analysis_parameters(); //Analysis paramaters
	} else {
		read_computation_parameters();
		read_graph_parameters();
	}
	
  //save base's size (to detect a change after computation)
	mainbasesize= game->main_base.size();
	checkbasesize= game->check_base.size();
	
  //signals that computation started
	game->trace_computation.set_stop_value(false);
	game->computation_result=global::Nim_unknown;
	computeRunning=true;
  //run thread
	emit isComputing(TRUE);
	if(computation_kind==global::Check || computation_kind==global::Complete_check) emit isChecking(TRUE);
	game->start(QThread::NormalPriority);
}

//Read user-defined parameters (position to compute)
void GlopWidget::read_computation_parameters() {
	game->parameter.spots_number=box_spots_number->value();
	game->parameter.use_given_position=given_position->isChecked();
	game->parameter.given_position=fromString( (box_given_position->text()).toStdString() );
	game->parameter.given_nimber=box_given_nimber->value();
}

//Read user-defined graph parameters (graph options)
void GlopWidget::read_graph_parameters() {
	game->check_graph.use_colors=box_use_colors->isChecked();
	game->check_graph.detail_level=box_detail->value();
	game->check_graph.use_graph=box_use_graph->isChecked();
	game->check_graph.min_lives=box_mlg->value();
}

//Read user-defined analysis parameters
void GlopWidget::read_analysis_parameters() {
  //number of lives
	game->lives_min_analysis=check_lives_min->isChecked();
	if(game->lives_min_analysis) game->lives_min=box_lives_min->value();
	game->lives_max_analysis=check_lives_max->isChecked();
	if(game->lives_max_analysis) game->lives_max=box_lives_max->value();
  //nimber
	game->nimber_min_analysis=check_nimber_min->isChecked();
	if(game->nimber_min_analysis) game->nimber_min=box_nimber_min->value();
	game->nimber_max_analysis=check_nimber_max->isChecked();
	if(game->nimber_max_analysis) game->nimber_max=box_nimber_max->value();
}

//stops computation
void GlopWidget::stop_computing() {
	game->trace_computation.set_stop_value(true);
}

//--------------------------------------------------------------------------
//Test functions
//--------------------------------------------------------------------------

//prints children of a given position
void GlopWidget::print_children() {
	textEdit_children->setPlainText("");	//clean text box
	Position the_pos( fromString( (box_given_position_2->text()).toStdString() ) );
	
	//compute children and fill the list
	the_pos.compute_children();	
	list<Position> result;
	set<Line>::iterator Li;
	for(Li=the_pos.children.begin();Li!=the_pos.children.end();Li++){
		result.push_back(Position(*Li));
	}
	
	//canonize children
	list<Position>::iterator Pi;
	for(Pi=result.begin();Pi!=result.end();Pi++){
		Pi->cano();	//children canonization
	}
	result.sort();		//needed for unique
	result.unique();	//removes all consecutive duplicate positions
	
	//print children in text box
	for(Pi=result.begin();Pi!=result.end();Pi++){
		textEdit_children->insertPlainText(QString(toString(Pi->r_vtx).c_str()));
		//print the nimber for the known children
		map<string, nimber>::const_iterator PNi=game->main_base.find(toString(Pi->r_vtx));
		if(PNi!=game->main_base.end()){	//the nimber is known in the database
			nimber the_nimber=PNi->second;
			textEdit_children->insertPlainText(QString(" "));
			textEdit_children->insertPlainText(QString("%1").arg(the_nimber,2,10));
		} else{				//the nimber could be known if there's several lands
			Pos_nim PN;
			PN.first=(Pi->r_vtx);
			PN.second=0;
			game->apply_known_nimbers(PN,true);
			if(PN.first.size()==1){	//every land is known, empty position : !
				nimber the_nimber=PN.second;
				textEdit_children->insertPlainText(QString(" "));
				textEdit_children->insertPlainText(QString("%1").arg(the_nimber,2,10));
			}
		}
		textEdit_children->insertPlainText(QString("\n"));
	}
}

//--------------------------------------------------------------------------
//Functions to trace computation
//--------------------------------------------------------------------------

//timer is activated when we receive true
//activate_timer is connected to isComputing
void GlopWidget::activate_timer(bool val) {
	if(val) {
		impulse->start(box_refresh->value()*100);
	} else {
		impulse->stop();
	}
}

//refresh widgets on each timer's impulse
void GlopWidget::trace_refresh() {
	//refresh widget displaying database's size
	refresh_base_size(game->parameter.computation_kind);
	
	if(game->parameter.computation_kind==global::Check && game->check_graph.use_graph) {
		refresh_base_size(global::Graph);
	}
	
	if (game->parameter.computation_kind!=global::Analysis) {
		widget_trace->update_display();
	}
}

//Refresh database size widget
void GlopWidget::refresh_base_size(int num_base) {
	unsigned long nb_pos=0; QString base_size;
	
	if(!game->base_lock.tryLock()) return;		//don't refresh size if database is not available
		if (num_base==global::Main) {
				nb_pos=game->main_base.size();
		}
		if (num_base==global::Check || num_base==global::Complete_check) {
				nb_pos=game->check_base.size();
		}
		if (num_base==global::Analysis){
				nb_pos=game->analysis_base.size();
		}
	game->base_lock.unlock();
	
	if (num_base==global::Main) {
		base_size = QString( "Main base : %1 " ).arg( nb_pos, 0, 10 );
		mainSizeLabel->setText(base_size);
	}
	if (num_base==global::Check || num_base==global::Complete_check) {
		base_size = QString( "Check base : %1 " ).arg( nb_pos, 0, 10 );
		checkSizeLabel->setText(base_size);
	}
	if (num_base==global::Analysis){
		base_size = QString( "Analysis base : %1 " ).arg( nb_pos, 0, 10 );
		analysisSizeLabel->setText(base_size);
	}
	if (num_base==global::Graph){
		unsigned long nb_pos_graph = game->check_graph.get_size();
		base_size = QString( "Graph base : %1 " ).arg( nb_pos_graph, 0, 10 );
		graphSizeLabel->setText(base_size);
	}
}

//Print Computation result
void GlopWidget::wait_result() {
	win_loss result;
	
	if(!game->game_lock.tryLock()) return;		//return if lock failed
		result=game->computation_result;
	game->game_lock.unlock();
	
	if (result==global::Nim_unknown) {
		//computation is still running
		return;
	}
	
	emit isComputing(FALSE); //signals computation ended
	emit isChecking(FALSE);
	computeRunning=false;
	
	//detect if main or check base changed
	if (mainbasesize != game->main_base.size()) {
		mainbaseisSaved=false;
	}
	if (checkbasesize != game->check_base.size()) {
		checkbaseisSaved=false;
	}
	
	//message displayed when a check computation failed
	int computation_kind=game->parameter.computation_kind;
	if ((computation_kind==global::Check || computation_kind==global::Complete_check) && game->check_failed) {
		QMessageBox::critical( this, "Check failed", QString::fromStdString(game->check_failure_msg));
		return;
	}
	
	//display message
	QString message;
	QString message_title="Result";
	if (computation_kind==global::Analysis) {
		message+="Analysis is completed.";
	}
	if (computation_kind==global::Complete_check) {
		if (result==global::Loss) {
			message+="Complete check stopped.";
		} else {
			message+="Database successfully checked";
		}
	}
	if (computation_kind==global::Main || computation_kind==global::Check){
		if (result==global::Computation_end) {
			message+="Computation stopped.";
		} else {
			//this is the normal case : a computation successfully ended
			if(!given_position->isChecked()) {
				//classic start position
				message+=QString("n=%1").arg(box_spots_number->value());
			} else {
				//free start position
				message+=box_given_position->text();
				if(box_given_nimber->value()!=0) {
					//add the nimber part (needed only if not null)
					message.prepend("(");
					message+=QString(" %1)").arg(box_given_nimber->value());
				}
			}
			
			if (result==global::Loss) {
				 message+=" is a losing position.";
			} else {
				message+=" is a winning position.";
			}
		}
	}
	
	QMessageBox::information (this, message_title, message, QMessageBox::Ok);
}


//--------------------------------------------------------------------------
//Read/write of databases
//--------------------------------------------------------------------------

//Purge Database
void GlopWidget::purgeBase(int num_base) {
	//confirm message
	QString msg_confirm;
	if(num_base==global::Main) {
		msg_confirm=QString("Do you really want to purge the main database ?");
	}
	if(num_base==global::Check) {
		msg_confirm=QString("Do you really want to purge the check database ?");
	}
	
	//ask confirmation
	//"No" button is default one
	int annulation=QMessageBox::question(this, "Confirmation", msg_confirm, "Yes", "No", QString(), 1, 1);
	
	//database is not deleted if not confirmed
	if(annulation) return;
	
	if(num_base==global::Main) {
		game->main_base.clear(); 	//main database
		mainbaseisSaved=true;
	}
	if(num_base==global::Check) {
		game->check_base.clear();	//check database
		checkbaseisSaved=true;
	}
	
	refresh_base_size(num_base);
}

//Save File Dialog Box. num_base parameter : which database to save
void GlopWidget::fileSaveAs(int num_base) {
	//database should not be empty
	bool baseIsEmpty=false;
	if(num_base==global::Main) {
			baseIsEmpty=game->main_base.empty();
	}
	if(num_base==global::Check) {
			baseIsEmpty=game->check_base.empty();
	}
	if(num_base==global::Analysis) {
			baseIsEmpty=game->analysis_base.empty();
	}
	if (baseIsEmpty) {
		QMessageBox::critical( this, "Warning", "There isn't any position to save.");
		return;
	}
	
	//dialog box
	QString filename = QFileDialog::getSaveFileName( this, "Save database", "", "*.spr");
	if ( !filename.isEmpty() ) {
		//add extension if there isn't
		if (!filename.endsWith(".spr")) {
			filename += ".spr";
		}
		
		setCursor( QCursor(Qt::WaitCursor) );
		write_file(filename, num_base);
		unsetCursor();
	}
}

//Open file Dialog Box
void GlopWidget::fileOpen() {
	QString filename = QFileDialog::getOpenFileName(this, "Add a database", "", "*.spr");
	if ( !filename.isEmpty() ) {
		setCursor( QCursor(Qt::WaitCursor) );
		read_file(filename);
		unsetCursor();
	}
}

//Save check graph
void GlopWidget::fileSaveGraph() {
	QString filename = QFileDialog::getSaveFileName(this, "Save check graph", "", "*.dot" );
	if ( !filename.isEmpty() ) {
		//add extension if there isn't
		if (!filename.endsWith(".dot")) {
			filename += ".dot";
		}
		
		//update graph parameters
		read_graph_parameters();
		
		//write graph file
		setCursor( QCursor(Qt::WaitCursor) );
		game->check_graph.write_file(filename);
		unsetCursor();
	}
}

//Write a base in the specified file
void GlopWidget::write_file( const QString &filename, int num_base ) {
	//open file
	QFile the_file( filename );
	if ( !the_file.open( QIODevice::WriteOnly ) ) return;
	
	//write database
	QTextStream text( &the_file );
	if(num_base==global::Analysis){	//analysis database id
		text << "[Pos+Nim+Nb_lives]" << "\n";
	}
	else{			//normal database or check database id
		text << "[Positions+Nimber]" << "\n";
	}
	
	map<string, nimber>::const_iterator SNi;
	QString chaine;
	
	//save normal database
	if(num_base==global::Main) {
		for(SNi=game->main_base.begin(); SNi!=game->main_base.end(); SNi++) {
			chaine=QString::fromStdString(SNi->first);
			text << chaine << " " << (int) (SNi->second) << "\n";
		}
		mainbaseisSaved=true;
	}
	//save check database
	if(num_base==global::Check) {
		for(SNi=game->check_base.begin(); SNi!=game->check_base.end(); SNi++) {
			chaine=QString::fromStdString(SNi->first);
			text << chaine << " " << (int) (SNi->second) << "\n";
		}
		checkbaseisSaved=true;
	}
	//save analysis database
	if(num_base==global::Analysis) {
		for(SNi=game->analysis_base.begin(); SNi!=game->analysis_base.end(); SNi++) {
			//print position and nimber
			chaine=QString::fromStdString(SNi->first);
			text << chaine << " " << (int) (SNi->second) << " ";
			
			//print position's number of lives
			text << nb_lives(fromString(SNi->first)) << "\n";
		}
	}
	
	//close file
	the_file.close();
}

//Add positions and their nimbers in the principal base, reading them from a file
void GlopWidget::read_file( const QString &filename ) {
	//open file
	QFile the_file( filename );
	if ( !the_file.open( QIODevice::ReadOnly ) ) return;
	
	QTextStream text( &the_file );
	
	//read database id
	if (!text.atEnd() ) {
		QString header;
		QString reference="[Positions+Nimber]";
		text >> header;
		if (header!=reference) {
			QMessageBox::critical( this, "Warning", "This file isn't a valid Positions+Nimber file.");
			the_file.close();
			return;
		}
	}
	
	//read data
	QString read_string;
	string str;
	int read_nim;
	nimber nim;
	
	map<string, nimber>::const_iterator SNi;
	
	text >> read_string;
	text >> read_nim;
	
	while ( !text.atEnd() ) {
		nim=read_nim;
		str=read_string.toStdString();
		
		SNi = game->main_base.find(str);
		if(SNi!=game->main_base.end()) {
			//already existing position. Check nimber consistency
			if (SNi->second != nim) {
				QMessageBox::critical( this, "Warning", "A position has a different nimber in the database and in the file." "Loading interrupted.");
				the_file.close();
				return;
			}
		} else {
			game->main_base[str]=nim; //add position to database
		}
		text >> read_string;
		text >> read_nim;
	}
	
	//close file
	the_file.close();
	
	//refresh principal database size widget
	refresh_base_size(global::Main);
}

//open the right popup
void GlopWidget::read_text_file(int num) {
	if(num==0) {
		WidgetPopup popup( this, QString(":/txt/help.html"), QString("Help"));
		popup.exec();
	} else if(num==1) {
		WidgetPopup popup( this, QString(":/txt/gpl.html"), QString("Licence"));
		popup.exec();
	} else if(num==2) {
		WidgetPopup popup( this, QString(":/txt/about.html"), QString("About"));
		popup.exec();
	}
}

//check that databases are saved when user closes the program
void GlopWidget::closeEvent(QCloseEvent *event) {
	if(mainbaseisSaved && checkbaseisSaved && !computeRunning) {
		//accept the close event (end of the program)
		event->accept();
		return;
	}
	
	//determine error message
	QString message;
	if(computeRunning) {
		message += "A computation is running.";
	} else
	if(!mainbaseisSaved) {
		message += "Main base has not been saved.";
	} else
	if(!checkbaseisSaved) {
		message += "Check base has not been saved.";
	}
	
	message += "\nDo you really want to quit?";
	
	//print message
	int annulation=QMessageBox::question(this, "Confirmation", message, "Yes", "No", QString(), 1, 1);
	
	if(annulation) {
		//user canceled the close : ignore the close event
		event->ignore();
	} else {
		//user chose to quit : accept the close event
		event->accept();
	}
}
