/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "graph.h"

#include <QtGui>

//clear graph databases
void Graph::initialize() {
	//delete any previous results
	nimbers_base.clear();
	index_base.clear();
	link_base.clear();
}

//return the size of index_base
unsigned long Graph::get_size() {
	QMutexLocker locker(&graph_lock);
	return index_base.size();
}

//get position's index (create a new one if needed)
pos_index Graph::get_index(const Pos_nim& pn) {
	//string representation of the position
	string pos=toString(pn.first);
	
	//test if the position is already stored
	map<string, pos_index>::iterator it;
	it = index_base.find(pos);
	
	pos_index result;
	
	if( it!=index_base.end() ) {
	  //index already exists
		result=it->second;
	} else {
	  //this position has no index, so we create one
		//current index (base size)
		result=index_base.size();
		
		//add result to index database
		index_base[pos]=result;
	}
	
	return result;
}

//update nimbers_base's informations with pn's data
void Graph::add_result(const Pos_nim& pn, win_loss pn_wl) {
	if( (!use_graph) || (nb_lives(pn.first)<min_lives) ) return;
	
	QMutexLocker locker(&graph_lock);
	
	//create (if necessary) a new index for pn
	get_index(pn);
	
	//string representation of the position
	string pos=toString(pn.first);
	
	//check if position already exists
	map<string, nimber_info>::iterator it;
	bool new_position=(nimbers_base.find(pos)==nimbers_base.end());
	
	//add result to win/loss database
	if(pn_wl==global::Loss){	//we know position's nimber
		nimbers_base[pos].first=global::Loss;
		nimbers_base[pos].second.clear();
		nimbers_base[pos].second.insert(pn.second);	//only one nimber is needed in the set
	}
	else{				//we don't know position's nimber
		if( !new_position ){
			//we only insert the wrong nimber if the right one isn't known yet
			if(nimbers_base[pos].first==global::Win) {
				nimbers_base[pos].second.insert(pn.second);
			}
		} else{
			//it's a new position, so we can insert the wrong nimber
			nimbers_base[pos].first=global::Win;
			nimbers_base[pos].second.insert(pn.second);
		}
	}
}

//add a link between the parent and the child in the link database
void Graph::add_link(const Pos_nim& parent, const Pos_nim& child) {
	if( (!use_graph) || (nb_lives(child.first)<min_lives) ) return;
	
	QMutexLocker locker(&graph_lock);
	
	//a move played in the nimber part : parent and child have the same position part,
	//so it's useless to create a link.
	if(parent.first==child.first) return;
	
	//get/create indexes for parent and child
	pos_index parent_index=get_index(parent);
	pos_index child_index=get_index(child);
	
	map<pos_index, set<pos_index> >::iterator it;
	set<pos_index> children_index_list;
	
	//check previous children of the parent
	it=link_base.find(parent_index);
	if(it!=link_base.end()) {
		children_index_list=it->second;
	}
	//add the new child to the list of children linked to the parent
	children_index_list.insert(child_index);
	
	//update the links starting from the parent
	link_base[parent_index]=children_index_list;
}

//stupid and slow algorithm to find a label with its index in index_base
// (note : the map object is not suited to perform search on the values instead of the keys)
map<string, pos_index>::iterator Graph::find_label(pos_index index) {
	map<string, pos_index>::iterator it;
	
	for(it=index_base.begin(); it!=index_base.end(); it++) {
		if (it->second==index) {
			return it;
		}
	}
	
	//we didn't find given index
	Error(QString("Element not found in Graph::find_label"), true);
	return it;	//useless, to avoid warning
}

//Write a file with "dot" format (for Graphviz)
void Graph::write_file( const QString &filename) {
	//delete multiple links (more information above the function definition).
	simplify_link_base();
	//open file
	QFile f( filename );
	if ( !f.open( QIODevice::WriteOnly ) ) return;
	
	QTextStream t( &f );
	
	//iterators
	map<pos_index, set<pos_index> >::iterator it_link;
	map<string, pos_index>::iterator it_index;
	map<string, nimber_info>::iterator it_nim;
	set<pos_index>::iterator it_set;
	
	//----------write the header--------------------------------------
	t << "digraph G {" << "\n\n";
	switch (detail_level){
	case 1:
		t << "edge [arrowhead=none];" << "\n";
		t << "node [shape=point];" << "\n"; break;
	case 2:
		t << "edge [arrowhead=vee];" << "\n";
		t << "node [fixedsize=true, width=0.55, height=0.55];" << "\n"; break;
	default:
		t << "edge [arrowhead=vee];" << "\n";
	}
	t << "\n";
	
	//----------write links-------------------------------------------
	if(detail_level==3){
	  //we write positions in the lexicographical order
		for(it_index=index_base.begin(); it_index!=index_base.end(); it_index++) {
			//find the place of this position in link_base
			it_link=link_base.find(it_index->second);
			//check that this position appears in link_base (as a parent)
			if(it_link!=link_base.end()){
				//print the parent
				t << print_position(it_index->second);
				//print the children
				t << "->{";
				for(it_set=it_link->second.begin(); it_set!=it_link->second.end(); it_set++) {
					//print the separator it it's not the first child
					if(it_set!=it_link->second.begin()) {
						t << ";";
					}
					//print the child
					t << print_position(*it_set);
				}
				t << "}\n";
			}
		}
	} else {
	  //we write numbers in the natural order ( 0, 1, 2 ...)
		for(it_link=link_base.begin(); it_link!=link_base.end(); it_link++) {
			//print the parent
			t << it_link->first;
			//print the children
			t << "->{";
			for(it_set=it_link->second.begin(); it_set!=it_link->second.end(); it_set++) {
				//print the separator it it's not the first child
				if(it_set!=it_link->second.begin()) {
					t << ";";
				}
				//print the child
				t << *it_set;
			}
			t << "}\n";
		}
	}
	t << "\n";
	
	//----------write labels------------------------------------------
	//we write labels in the natural order ( 0, 1, 2 ...)
	int last_index=index_base.size();
	
	for(int i=0;i<last_index; i++) {
		it_index=find_label(i);
		it_nim=nimbers_base.find(it_index->first);
		if (it_nim==nimbers_base.end()){
			Error(QString("Element not found in Graph::write_file"), true);
			return;
		}
		
		string position=it_index->first;
		char type_pos=Position(fromString(position)).position_type();
		
		//print position, or index (depending on the detail_level)
		if(detail_level==3){
			t << print_position(i);
		} else{
			t << i;
		}
		t<< "\t" << "[";
		
		//shape according to the type of position (Loss/Win/several lands)
		if(detail_level>1){
			t << "shape=";
			if(type_pos==global::Single_land){
				if(it_nim->second.first==global::Loss) {
					t << "ellipse, ";
				} else {
					t << "none, ";
				}
			}
			else {
				t << "box, ";
			}
		}
		
		//colors (optional, same information as shape but more visible)
		if(use_colors){
			t << "style=filled, color=";
			if(type_pos==global::Single_land){
				if(it_nim->second.first==global::Loss) {
					t << "red";
				} else {
					t << "blue";
				}
			}
			else {
				t << "yellow";
			}
		}
		
		t << "]";
		
		//print position as a comment
		if(detail_level!=3){
			t << " //" << print_position(i);
		}
		
		t << "\n";
	}
	t << "\n";
	
	//----------create subgraphs--------------------------------------
	//We use subgraphs to print to the same rank positions with the same number of lives
	map< int, list<pos_index> > lives_levels;
	map< int, list<pos_index> >::iterator Level_i;
	map< int, list<pos_index> >::reverse_iterator Level_ri;
	list<pos_index>::iterator Index_j;
	
	//put the right index in the right list, depending on the number of lives of its position
	for(int i=0;i<last_index; i++) {
		it_index=find_label(i);
		lives_levels[nb_lives(fromString(it_index->first))].push_back(i);
	}
	
	//print a vertical line in the graph to indicate the number of lives of positions in this rank
	for(Level_ri=lives_levels.rbegin();Level_ri!=lives_levels.rend();Level_ri++){
		if(Level_ri!=lives_levels.rbegin()){
			t << "->";
		}
		t << "L" << Level_ri->first;
	}
	t << ";\n\n";
	
	//print the subgraphs
	for(Level_i=lives_levels.begin();Level_i!=lives_levels.end();Level_i++){
		t << "{rank=same; ";
		for(Index_j=Level_i->second.begin();Index_j!=Level_i->second.end();Index_j++){
			if(detail_level==3){
				t << print_position(*Index_j) << "; ";
			} else{
				t << *Index_j << "; ";
			}
		}
		t << "L" << Level_i->first << ";}\n";
	}
	
	//-----end of file-----
	t << "\n}\n";
	
	//close file
	f.close();
}

//In the link base, if P is a position (and a single land) whose nimber is unknown,
//but if we know that this nimber is neither n1 nor n2 ... nor nk,
//we only need one link P->P1 where P1's nimber is n1, one link P->P2.
//This function deletes multiple links from such a position to children with same nimbers.
//Such multiple links can occur because position's whose nimber is unknown aren't stored by the program,
//so it sometimes happens that the program checks that P's nimber is not n with a child C1 at first,
//and later check that P's nimber is not n with another child C2.
void Graph::simplify_link_base() {
	map<pos_index, set<pos_index> >::iterator it_link;
	set<pos_index>::iterator Index_c,Index_tmp;
	map<string, pos_index>::iterator it_index;
	map<string, nimber_info>::iterator it_nim_p,it_nim_c;
	set<nimber>::iterator Ni;
	
	//test any position linked to another, the position P's index is "it_link->first"
	for(it_link=link_base.begin(); it_link!=link_base.end(); it_link++) {
		it_index=find_label(it_link->first);
		it_nim_p=nimbers_base.find(it_index->first);
		if(it_nim_p==nimbers_base.end()){
			Error(QString("Element not found in Graph::simplify_link_base (1)"), true);
		}
		//nothing to do if the position is a Loss
		if((it_nim_p->second).first==global::Win){
			set<nimber> set_nim=(it_nim_p->second).second;
			//test any nimber that is known not to be P's nimber
			for(Ni=set_nim.begin();Ni!=set_nim.end();Ni++){
				bool one_child_found=false;
				Index_c=(it_link->second).begin();	//iterator on the index of the children of P
				//check any child linked to P
				while(Index_c!=(it_link->second).end()){
					//iterator on the child of P which is studied
					it_nim_c=nimbers_base.find(find_label(*Index_c)->first);
					//security
					if(it_nim_c==nimbers_base.end()){
						Error(QString("Element not found in Graph::simplify_link_base (2)"), true);
					}
					//check if the nimber of the child is equal to *Ni
					if(*((it_nim_c->second).second.begin())==*Ni){
						if(one_child_found){	//not the first child with this nimber : erase it
							Index_tmp=Index_c; Index_tmp++;
							(it_link->second).erase(Index_c);
							Index_c=Index_tmp;
						} else {	//first child with this nimber : do nothing
							one_child_found=true;
							Index_c++;
						}
					} else {
						Index_c++;
					}
				}
			}
		}
	}
}

//print the position whose index is i
QString Graph::print_position(pos_index i){
	map<string, pos_index>::iterator it_index=find_label(i);
	map<string, nimber_info>::iterator it_nimber=nimbers_base.find(it_index->first);
	if(it_nimber==nimbers_base.end()){
		Error(QString("Element not found in Graph::print_position"), true);
	}
	QString result;
	result.append((char) 34);
	//print position
	result.append(QString::fromStdString(it_index->first));
	//warn if nimber is unknown
	if(it_nimber->second.first==global::Win) {
		result.append(" not");
	}
	//print nimber(s)
	set<nimber>::iterator Ni;
	for(Ni=it_nimber->second.second.begin();Ni!=it_nimber->second.second.end();Ni++){
		result.append(" ");
		result.append(QString::number(*Ni));
	}
	result.append((char) 34);
	return result;
}
