/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "land.h"

//override terminal_vertex virtual function
Vertex Land::terminal_vertex() {
	return global::End_Land;
}


//-------------------------------------------------------------------------------------------------
// Reduction
//-------------------------------------------------------------------------------------------------

//delete vertices of degree 3
void Land::delete_dead_vertices() {
//note : there is at most two dead vertices. It could be used to optimize this function more
	//-----delete generic vertex '3'---------------------
	r_vtx.remove(3);
	
	//-----delete upper case vertices of degree 3--------
	//vector to remember the number of times an upper-case letter appears
	//size of vector is equal to the number of possible upper-case letters
	vector<int> vertex_degree(global::Last_up_vtx-global::First_up_vtx+1);

	//initialization, convention : 0 means the letter is not used
	fill(vertex_degree.begin(), vertex_degree.end(), 0);
	
	//note : index in the vector = vertex name - global::First_up_vtx
	//compute the degree of each vertex
	Line::iterator s;
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		if(is_upper_case(*s)) {
			vertex_degree[*s-global::First_up_vtx]+=1;
		}
	}
	
	//delete dead vertices (degree 3)
	Vertex i;
	for(i=global::First_up_vtx; i<=global::Last_up_vtx; i++) {
		if(vertex_degree[i-global::First_up_vtx]==3) {
			r_vtx.remove(i);
		}
	}
}

//delete adjacent vertices
void Land::delete_adjacent_vertices() {
	//s1 and s2 are used to go through the list, s2 is always equal to s1's next element
	//s_begin is used to remember the first vertex of a Boundary
	Line::iterator s_begin=r_vtx.begin(), s1=r_vtx.begin(), s2=r_vtx.begin();
	
	//s2 initialization
	if (s2 != r_vtx.end() ) { //security test
		s2++;
	}
	
	bool increment_needed=true; //indicate if we need to increment s1 and s2 at the end of the loop
	
	while(s2 != r_vtx.end()) {
		//----- detect new boundary ------------
		//update s_begin (first vertex of the boundary) when we enter a new boundary
		//note : a new boundary is "detected" when a "non-real" vertex is followed by a "real" one
		if (!is_real_vtx(*s1) && is_real_vtx(*s2)) {
			s_begin=s2;
		}
		
		//---------- test if s1 is adjacent ------
		//adjacence has meaning only for letters
		if( is_letter_vtx(*s1) ) {
			//adjacence between last and first real vertex of a boundary longer than one vertex
			if(*s2==global::End_Boundary && s_begin != s1 && *s_begin == *s1) {
				s1=r_vtx.erase(s1); //erase and increment s1
				s2++;
				increment_needed = false;
			}
			
			//adjacence between two consecutive vertices
			if(*s1 == *s2) {
				// caution : we must erase s2, because s1 and s_begin can refer to the same vertex
				// and s_begin should never be erased
				s2=r_vtx.erase(s2); //erase and increment s2
				increment_needed = false;
			}
		}
		
		//---------- increment s1 and s2 if necessary -------
		if (increment_needed) {
			s1++;
			s2++;
		}
		increment_needed=true;  //initialization for next loop
	}
}

//delete empty boundaries
void Land::delete_empty_boundaries() {
	//s2 is always equal to the element immediatly after s1
	Line::iterator s1=r_vtx.begin(), s2=r_vtx.begin();
	if (s2 != r_vtx.end()) {
		s2++;
	}
	
	while(s2 != r_vtx.end()) {
		if (s1==r_vtx.begin() && *s1==global::End_Boundary) {
			//empty boundary in the beginning of the vertex list : .
			s1=r_vtx.erase(s1);	//erase s1 and place it on next vertex
			s2++;			//place s2 on next vertex
		} else if (*s2==global::End_Boundary && (*s1==global::End_Boundary || *s1==global::End_Domain)) {
			//empty boundary in other cases :  .. or }.
			//boundary is empty : we just delete the 'end of boundary' character
			//erase s2 and place it on next vertex (s1 stays at the same vertex)
			s2=r_vtx.erase(s2);
		} else {
			//incrementation only if we didn't delete any empty boundary
			s1++; s2++;
		}
	}
}

//delete dead domains (domain list is supposed to be uptodate)
void Land::delete_dead_domains() {
	list<Domain>::iterator Di;
	
	Di=r_str.begin();
	while( Di!=r_str.end()) {
		if(Di->dead()) {
			//erase the dead domain and place Di on the next domain
			Di=r_str.erase(Di);
		} else {
			//just place Di on the next domain
			Di++;
		}
	}
}

//merge boundaries in domains (domain equivalence theory)
void Land::merge_boundaries(){
	list<Domain>::iterator Di;
	for(Di=r_str.begin(); Di!=r_str.end(); Di++) {
		Di->merge_boundaries();
	}
}

//generic name '2' used for :
// - upper case vertices which occur only one time in the land
// - lower case vertices which occur only one time in a boundary : this pathological case
//happens after delete_adjacent_vertices(), example :
//a position : 1aAa.}2A.}]!, a child : 1aAa.}3*A.}3A*.}]!, after delete_dead_vertices : 1aa.}*.}*.}],
//after delete_adjacent_vertices : 1a.}*.}*.}], etc... this function tranforms 1a. into 12.
void Land::apply_generic_vertex() {
	Line::iterator p, q, s;
	p=r_vtx.begin(); q=r_vtx.end();
	
	//this map stores all the lower-case vertices that appears only once in a given boundary
	//for a given letter, we remember its position in the boundary (the iterator) to achieve quick renames
	map<Vertex, Line::iterator> low_is_generic;
	map<Vertex, Line::iterator>::iterator it;
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		//-----upper case or temporary vertices-----
		if( is_upper_case(*s) || is_temp_vtx(*s) ) {
			//rename the vertex with two if it appears only once in the land
			if(count(p,q,*s)==1) {
				*s=2;
			}
		}
		
		//-----lower case vertices detection--------
		if(is_lower_case(*s)) {
			//if vertex already exists, then delete it from map, else add it
			it=low_is_generic.find(*s);
			if(it!=low_is_generic.end()) {
				low_is_generic.erase(it);
			} else {
				//store (*s, s)
				low_is_generic[*s]=s;
			}
		}
		
		//-----rename lower case vertices (when boundary ends) -----
		if(*s==global::End_Boundary) {
			//rename with "2" every vertex of the map
			for(it=low_is_generic.begin(); it!=low_is_generic.end(); it++) {
				*(it->second)=2;
			}
			
			//clear map
			low_is_generic.clear();
		}
	}
}

//reduction of land vertex list representation
//note : the order of the different treatments is very important
void Land::reduce(){
	change_vertex();
		delete_dead_vertices();
		delete_adjacent_vertices();
		delete_empty_boundaries();
	update_representation();
	
	change_structure();
		delete_dead_domains();
		merge_boundaries();
		distinguish_low_up();
	update_representation();
	
	change_vertex();
		apply_generic_vertex();    //use generic name '2'
}


//-------------------------------------------------------------------------------------------------
// Canonization
//-------------------------------------------------------------------------------------------------

//rename upper-case and temporary vertices in their order of occurence
//with global::First_up_vtx as the first letter
void Land::rename_in_order() {
	vector<int> new_names(global::Last_vtx+1);  //vertices new names
	
	//initialization : 0 means here "new name is not yet known"
	fill(new_names.begin(), new_names.end(), 0);
	
	Line::iterator s;
	int next_letter=global::First_up_vtx; //next unused name (we begin at First_up_vtx)
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		//check if vertex is a temp/upper-case letter
		if(is_upper_case(*s) || is_temp_vtx(*s)) {
			//check if a new name was already given to this vertex
			if (new_names[*s]==0) {
				new_names[*s]=next_letter;	//give a new name (first available letter)
				*s=next_letter;
				next_letter++;			//increment the first available letter
			} else {
				*s=new_names[*s];		//use the already given new name
			}
		}
	}
	
	//security on the number of used letters
	next_letter--; //last used upper-case letter
	if (next_letter>global::Last_up_vtx) {
		Error(QString("Land::rename_in_order : upper-case letters capacity overflow"), true);
	}
}

//sort land's representation
void Land::sort() {
	//sort each domain
	list<Domain>::iterator Di;
	for(Di=r_str.begin(); Di!=r_str.end(); Di++) {
		Di->sort();
	}
	
	//lexicographical sort of domains
	r_str.sort();
}

//canonization of land
//basically, this is two (rename+sort)
void Land::cano() {
	change_vertex();
		rename_in_order();
	update_representation();

	change_structure();
		sort();
		
		if (exist_lower(r_vtx)) {
			rename_low_in_order();
		}
	update_representation();
	
	change_vertex();
		rename_in_order();
	update_representation();

	change_structure();
		sort();
	update_representation();
}


//--------------------------------------------------------------------------
//split land in new independent ones
//--------------------------------------------------------------------------

//test if two lists of vertices have a common temp/upper-case vertex
//when two lists are "linked" it means they are part of the same land
bool linked(const Line& a1, const Line& a2) {
	Line::const_iterator s1, s2;
	
	for(s1=a1.begin(); s1!=a1.end(); s1++) {
		if(is_upper_case(*s1) || is_temp_vtx(*s1)) {
			for(s2=a2.begin(); s2!=a2.end(); s2++) {
				if (*s1 == *s2) {
					//we found a common temp/upper-case vertex
					return true;
				}
			}
		}
	}
	
	//there is no common vertex
	return false;
}

//split algorithm : try to split the land in independent ones
//the result is stored in the land vertex representation
// we return a boolean indicating if split occured or not
bool Land::split() {
	update_representation();
	list<Line> indep_land;		//list of independent lands (split function result)
	
	list<Domain>::iterator Di;
	list<Line>::iterator Lj;	//iterator on the list of independent lands
	
	Line linked_vertices;		//list of vertices linked to the current domain
	list<Line> relative_indep_land;	//list of lands independent from the current domain
	bool new_land;			//indicate if current domain is a new independant land or not
	
	//loop on the domains
	for(Di=r_str.begin(); Di!=r_str.end(); Di++) {
		//we separate all already found lands into two categories :
		linked_vertices.clear();	// - linked to the current domain (current domain's vertex "appears")
		relative_indep_land.clear(); 	// - independent from the current domain
		new_land=true;
		
		for(Lj=indep_land.begin(); Lj!=indep_land.end(); Lj++) {
			if (linked(*Lj, Di->r_vtx)) {
				//add this land to the "tied list"
				linked_vertices.insert(linked_vertices.end(), Lj->begin(), Lj->end());
				new_land=false;
			} else {
				//this domain doesn't interfere with land's freedom
				//so we add the land in the list of free lands
				relative_indep_land.push_back(*Lj);
			}
		}
		
		if (new_land) {
			//create a new land (domain is independent from all already existing ones)
			relative_indep_land.push_back(Di->r_vtx);
		} else {
			//add the domain to "the tied list"
			linked_vertices.insert(linked_vertices.end(), Di->r_vtx.begin(), Di->r_vtx.end());
		}
		
		//update the complete list of independent land
		indep_land.clear();
		indep_land.insert(indep_land.begin(), relative_indep_land.begin(), relative_indep_land.end());
		if ( !linked_vertices.empty() ) {
			indep_land.push_back(linked_vertices);
		}
	}
	
	//add 'end of land' characters
	for(Lj=indep_land.begin(); Lj!=indep_land.end(); Lj++) {
		Lj->push_back(global::End_Land);
	}
	
	//result is saved in the vertices list of the land
	//this is an exceptional case where we use vertices list to store more than one land
	if(indep_land.size()>1) {
		//split occured
		r_vtx.clear();
		for(Lj=indep_land.begin(); Lj!=indep_land.end(); Lj++) {
			r_vtx.insert(r_vtx.end(), Lj->begin(), Lj->end());
		}
		return true;
	} else {
		//split didn't occur
		return false;
	}
}


//--------------------------------------------------------------------------
//Children of the land
//--------------------------------------------------------------------------
//this function is similar to Position::compute_children()
//specific code is clearly indicated

void Land::compute_children() {
	update_representation();
	children.clear();
	
	//specific code - compared to Position::compute_children()
	init_free_up_vtx();		//initialization of the first available upper-case letter
	
	list<Domain>::iterator it;
	list<Domain>::iterator it2;
	//end of specific code
	
	set<Line>::iterator it_children;
	
	Line l_unused; //vertex representation of the list of unused sub-structures
	Line l_tmp;
	
	//compute children of each sub-structure
	//a child is a sub-structure child completed with the unused other substructures
	
	//loop on the substructures
	for(it=r_str.begin(); it!=r_str.end(); it++) {
		//compute substructure's children
		it->compute_children();
		
		//list of unused sub-structures
		l_unused.clear();
		for(it2=r_str.begin(); it2!=r_str.end(); it2++) {
			if(it2!=it) {
				//this is an unused substructure -> we add it to l_unused
				l_unused.insert(l_unused.end(), it2->r_vtx.begin(), it2->r_vtx.end());
			}
		}
		
		//we add the list of unused sub-structures to each child of the used sub-structure
		//loop on the sub-structure's children
		for(it_children=it->children.begin(); it_children!=it->children.end(); it_children++) {
			l_tmp.clear();                                                 //delete temporary list
			l_tmp.insert(l_tmp.begin(), l_unused.begin(), l_unused.end()); //add list of unused sub-structures
			l_tmp.insert(l_tmp.end(), it_children->begin(), it_children->end());   //add child's representation
			l_tmp.push_back(terminal_vertex());                            //add 'end of structure'
			
			children.insert(l_tmp);
		}
	}
}

//Estimate number of children in the land
unsigned int Land::estimate_children_number() {
	update_representation();
	
	unsigned int estimation=0;
	list<Domain>::iterator Di;
	
	//sum of estimation in each "domain"
	for(Di=r_str.begin(); Di!=r_str.end(); Di++) {
		estimation += Di->estimate_children_number();
	}
	
	return estimation;
}

//-------------------------------------------------------------------------
//lower-upper case distinction
//--------------------------------------------------------------------------

//Rename upper-case letters in lower-case ones (when needed)
void Land::distinguish_low_up() {
	list<Domain>::iterator Di;
	for(Di=r_str.begin(); Di!=r_str.end(); Di++) {
		Di->distinguish_low_up();
	}
}

//rename lower-case letter in their order of appearance
void Land::rename_low_in_order() {
	
	list<Domain>::iterator Di;
	for(Di=r_str.begin(); Di!=r_str.end(); Di++) {
		Di->rename_low_in_order();
	}
}

//find first available upper-case letter
//this value can be computed only at the Land object level, but is needed by Domain::compute_children
void Land::init_free_up_vtx() {
	update_representation();
	
	//find first available upper-case letter
	Line::iterator s;
	Vertex free_vtx=global::First_up_vtx; //next available upper-case Vertex
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		//test upper-case and if it's greater than the current available Vertex
		if(is_upper_case(*s) && *s>=free_vtx) {
			free_vtx=*s+1; //*s is used. We need a new value for current available letter
		}
	}
	
	//error case
	if (free_vtx>global::Last_up_vtx) {
		Error(QString(" Land::init_free_up_vtx : upper-case letters capacity overflow"), true);
	}
	
	//store result in a static variable available everywhere
	Representation::free_upper_vtx=free_vtx;
}
