/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "position.h"


//--------------------------------------------------------------------------
//override terminal_vertex virtual function
//--------------------------------------------------------------------------
Vertex Position::terminal_vertex() {
	return global::End_Position;
}


//--------------------------------------------------------------------------
//canonization of the position
//--------------------------------------------------------------------------
void Position::cano() {
	list<Land>::iterator Li;
	
	//--------- step 1 : reduce lands ----------
	update_representation();
	change_structure();
	for(Li=r_str.begin(); Li!=r_str.end(); Li++) {
		Li->reduce();
	}
	
	//--------- step 2 : try to split lands into independent new ones ----------
	bool split_occured=false;
	for(Li=r_str.begin(); Li!=r_str.end(); Li++) {
		if(Li->split()) {
			split_occured = true;
		}
	}
	
	if(split_occured) {
		//important note : when split occurs, we store several lands in the initial Land object
		//we first compute position's vertex representation
		//then we compute the "real" structure representation (one Land object for each land)
		change_structure();        //indicate that the structure representation has changed
		update_representation();   //update vertex list
		change_vertex();           //pretend vertex list has changed
		update_representation();   //update structure list
	}
	
	//---------- step 3 : canonize lands ----------
	change_structure();
	Li=r_str.begin();
	while(Li!=r_str.end()) {
		//canonization
		Li->cano();
		
		//delete the land if dead (r_vtx : "]")
		if(Li->r_vtx.size()==1){
			Li=r_str.erase(Li);
		} else {
			Li++;
		}
	}
	
	//lexicographical sort of land
	r_str.sort();
	update_representation();
}

//Return the list of lands considered as positions
// returned result : a list of vertex representations
list<Line> Position::list_lands() {
	list<Line> result;	//Line=vertex representation
	Line land_tmp;
	Line::const_iterator s;
	
	for(s=r_vtx.begin(); s!=r_vtx.end(); s++) {
		if (*s==global::End_Position) {
			//this character is just ignored
		} else if(*s==global::End_Land) {
			//add 'end of land' and 'end of position' character
			land_tmp.push_back(global::End_Land);
			land_tmp.push_back(global::End_Position);
			
			//add the land to the result list
			result.push_back(land_tmp);
			land_tmp.clear();
		} else {
			//add one character to the current land's vertex list
			land_tmp.push_back(*s);
		}
	}
	return result;
}

//Children of the position
//result is saved in the class member "children"
//this function is similar to Land::children
//Note : this function is always called on a single land in the Game class, but can be called
//on several lands in the "Children" tab
void Position::compute_children() {
	update_representation();
	children.clear();
	
	//specific code - compared to Land::compute_children()
	list<Land>::iterator it;
	list<Land>::iterator it2;
	//end of specific code
	
	set<Line>::iterator it_children;
	Line l_unused; //vertex representation of the list of unused sub-structures
	Line l_tmp;
	
	//compute children of each sub-structure
	//a child is a sub-structure child completed with the unused other substructures
	
	//loop on the substructures
	for(it=r_str.begin(); it!=r_str.end(); it++) {
		//compute substructure's children
		it->compute_children();
		
		//list of unused sub-structures
		l_unused.clear();
		for(it2=r_str.begin(); it2!=r_str.end(); it2++) {
			if(it2!=it) {
				//this is an unused substructure -> we add it to l_unused
				l_unused.insert(l_unused.end(), it2->r_vtx.begin(), it2->r_vtx.end());
			}
		}
		
		//we add the list of unused sub-structures to each child of the used sub-structure
		//loop on the sub-structure's children
		for(it_children=it->children.begin(); it_children!=it->children.end(); it_children++) {
			l_tmp.clear();                                                 //delete temporary list
			l_tmp.insert(l_tmp.begin(), l_unused.begin(), l_unused.end()); //add list of unused sub-structures
			l_tmp.insert(l_tmp.end(), it_children->begin(), it_children->end());   //add child's representation
			l_tmp.push_back(terminal_vertex());                            //add 'end of structure'
			
			children.insert(l_tmp);
		}
	}
}

//Return the type of position
//It depends on the number of land substructures
char Position::position_type() {
	//counts the number of land
	int nb_land=count(r_vtx.begin(), r_vtx.end(), global::End_Land);
	if(nb_land==0) {
		return global::Empty_position;
	} else if(nb_land==1) {
		return global::Single_land;
	} else {
		return global::Multi_lands;
	}
}

//Estimate number of possible children of the position
unsigned int Position::estimate_children_number() {
	update_representation();
	
	unsigned int estimation=0;
	list<Land>::iterator Li;
	
	//sum of estimation in each land
	for(Li=r_str.begin(); Li!=r_str.end(); Li++) {
		estimation += Li->estimate_children_number();
	}
	
	return estimation;
}
