/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "global.h"

#ifndef REPRESENTATION_H
#define REPRESENTATION_H

// Virtual base class representing a game object, essentially a vertex list
// All the other classes (Boundary, Domain, Land, Position) derive from Representation

class Representation {
	//operator<< to print vertex list with cout
	friend ostream& operator<<(ostream&, const Representation&);
	
  public:
	Line r_vtx;           //vertex list representation
	set<Line> children;   //set of children (successors)
	
	Representation(const Line&);	//initialization with a vertex list
	Representation() {};
	virtual ~Representation() {};
	
	void reset(const Line&);
	
	//note : this two members have meaning only in the Structure class
	//but we define them here for simplicity of reset() and update_structure() code
	bool vertex_is_uptodate;     //indicates if the vertex list is uptodate
	bool structure_is_uptodate;  //indicates if the structure list is uptodate
	
	//purely virtual function returning the special "end of this structure" vertex
	//this function is overrided in each derived class
	virtual Vertex terminal_vertex() = 0;
	
	//comparison operator (needed by STL objects, in particular list::sort, set and map)
	bool operator<(const Representation&) const;
	bool operator==(const Representation&) const;
	
	// static variable (global to the class) used in the lower-upper distinction algorithm
	static Vertex free_upper_vtx;	//first non-used upper-case letter
};


//-------------------------------------------------------------------------------
//Structure is the base class for "complex structures", ie Domain, Land, Position
//-------------------------------------------------------------------------------

//this class is a template, because we speak of "a structure of something"
//for example : Domain is a structure of Boundaries

template <class T>
class Structure : public Representation {
	
  public:
	list<T> r_str;					//structure representation (list of sub-objects)
	
	Structure(const Line& init) : Representation(init) {};
	Structure() {};
	
	void update_vertex();			//update vertex list (using the structure list)
	void update_structure();		//update structure list (using the vertex list)
	void update_representation();	//use the appropriate list to update the other
	
	//these functions must be called before making any change to a list
	//this is a secured way to remember which list is uptodate and which is not
	void change_vertex();			//before changing the vertex list
	void change_structure();		//before changing the structure list
};


//--------------------------------------------------------------------------
//implementation of Structure in this .h file because it's a template
//--------------------------------------------------------------------------

//update the list of vertex from the list of structures
template <class T>
void Structure<T>::update_vertex() {
	typename list<T>::const_iterator Ti;
	
	//delete old list of vertices
	r_vtx.clear();
	
	//add the representation of the sub-structures to the list of vertex
	for(Ti=r_str.begin(); Ti != r_str.end(); Ti++) {
		r_vtx.insert(r_vtx.end(), Ti->r_vtx.begin(), Ti->r_vtx.end());
	}
	
	//add an 'end of structure' character. For example, for domain : End_Domain
	r_vtx.push_back(terminal_vertex());
	
	vertex_is_uptodate=true;
}


//update structure (the list of sub-structures) from the list of vertices
//To save runtime, this function doesn't destroy old sub-structures list
//but tries instead to update existing sub-structures, creating/deleting objects only if needed

//caution : be very careful to never suppose an object is a "new one"
//(particularly with children class member for example)

template <class T>
void Structure<T>::update_structure() {
	//we need at least one empty T object
	Line r_empty;
	T T_empty(r_empty);
	
	// T object end vertex is T::terminal_vertex()
	Vertex end_sub_str=T_empty.terminal_vertex();
	
	//s_begin and s_end are the beginning and the end vertex of a T structure
	//s is the current vertex
	Line::const_iterator s_begin, s_end, s;
	
	typename list<T>::iterator Ti;	//iterator on the structure list
	
	Ti=r_str.begin();
	s_begin=r_vtx.begin();
	
	//simple loop on the vertex list searching T objects end vertex
	for(s=r_vtx.begin(); s != r_vtx.end(); s++) {
		if (*s==end_sub_str) {
			//we found an end of T object
			
			//Ti must always refer to a valid structure in the structure list
			if(Ti==r_str.end()) {
				Ti=r_str.insert(r_str.end(), T_empty );
			}
			
			//place s_end after the last vertex of the T object
			s_end=s;
			s_end++;
			
			//store the found T object (vertices between s_begin and s_end) in Ti
			Ti->r_vtx.clear();
			Ti->r_vtx.insert(Ti->r_vtx.end(), s_begin, s_end);  //note : s_end is not inserted
			Ti->structure_is_uptodate=false;
			Ti->vertex_is_uptodate=true;
			
			//prepare next step
			s_begin=s_end;    //the next T object begins at s_end
			Ti++;             //place Ti on the next structure of the list 
		}
	}
	
	//destroy unused sub-structures
	if(Ti!=r_str.end()) {
		r_str.erase (Ti, r_str.end());
	}
	
	structure_is_uptodate=true;
}


//use the appropriate list to update the other
//after a call to this function, both representations are uptodate

template <class T>
void Structure<T>::update_representation() {
	if (!vertex_is_uptodate) {
		update_vertex();
	}
	
	if(!structure_is_uptodate) {
		update_structure();
	}
}


//--------------------------------------------------------------------------
//functions that must be called to indicate a list is going to be changed
//--------------------------------------------------------------------------
//if we change a list, the other one isn't uptodate any more

template <class T>
void Structure<T>::change_vertex() {
	if (!vertex_is_uptodate) {
		Error(QString("Structure<T>::change_vertex : Attempt to change r_vtx, which is not uptodate"), true);
	}
	structure_is_uptodate=false;
}

template <class T>
void Structure<T>::change_structure() {
	if (!structure_is_uptodate) {
		Error(QString("Structure<T>::change_structure : Attempt to change r_str, which is not uptodate"), true);
	}
	vertex_is_uptodate=false;
}

#endif
