/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "trace.h"
#include <QtGui>

//--------------------------------------------------------------------------
//Communication with "Game" class
//--------------------------------------------------------------------------

//initialize the trace
void Trace::initialize() {
	QMutexLocker locker(&trace_lock);
	
	current_level=0;
	parent_level=-1;
	signal=0;		//no signal
	signal_child_number=-1;
	reverse_zapping=false;
	traced_levels.clear();
	stop_computing=false;	//"false" allows the computing to be launched without graphical interface (from -t option)
}

//create a new level
void Trace::create_level(bool for_nimber) {
	QMutexLocker locker(&trace_lock);
	
	current_level++;		//increment number of levels
	
	//add an empty level to the list of levels
	Level empty_level;
	empty_level.nimber_level=for_nimber;
	empty_level.nb_alive_pos=0;
	traced_levels.push_back(empty_level);
}

//add only one position-nimber to the bottom-level
void Trace::add_one_pos_nim(const Pos_nim& pn) {
  //security to avoid a possible crash (if the levels' list is empty, rbegin crashes)
	if (traced_levels.empty()) {
		Error(QString("Trace::add_one_pos_nim : no level available"), false);
		return;
	}
  //access the last level with the standard function rbegin
	list<Level>::reverse_iterator last_level;
	last_level=traced_levels.rbegin();
  //parameters
	PositionTrace next_pos;
	next_pos.r_vtx=pn.first;
	next_pos.nim=pn.second;
	next_pos.status_unknown=true;	//position is "alive"
	next_pos.nb_lives=nb_lives(pn.first);	//compute number of lives
  //add position and information to the last level
	(last_level->traced_positions).push_back(next_pos);
	last_level->nb_alive_pos++;
}

//add a position-nimber to the bottom-level
void Trace::add(const Pos_nim& pn) {
	QMutexLocker locker(&trace_lock);
	
	add_one_pos_nim(pn);
}

//add a list of position-nimber to the bottom-level
void Trace::add(const list<Pos_nim> & l_pn) {
	QMutexLocker locker(&trace_lock);
	
	list<Pos_nim>::const_iterator Pi;
	
	for(Pi=l_pn.begin(); Pi!=l_pn.end(); Pi++) {
		add_one_pos_nim(*Pi);
	}
}


//function called by Game::compute_win_loss when studied position is changed
//the two possible reasons are : a child was found "winning" or a different branch was chosen by user
//this function changes the index of the studied position of the lowest level
void Trace::change_studied_pos(int pos_index){
	QMutexLocker locker(&trace_lock);
	
	//security to avoid possible crash of rbegin
	if (traced_levels.empty()) {
		Error(QString("Trace::change_studied_pos : no level available"), false);
		return;
	}
	
	(traced_levels.rbegin())->studied_pos_index=pos_index;
}

//called by Game::compute_win_loss when a child status has been found "winning"
//we keep trace of child status
void Trace::known_child(int pos_index){
	QMutexLocker locker(&trace_lock);
	
	//security to avoid any possible crash (this one is probably a paranoid one)
	if (traced_levels.empty()) {
		Error(QString("Trace::known_child : no level available"), false);
		return;
	}
	
	//Go to the correct level
	list< Level >::iterator Li=traced_levels.begin();
	int i;
	for(i=1;i<current_level && Li!=traced_levels.end();i++){
		Li++;
	}
	//update number of alive positions
	Li->nb_alive_pos--;
	
	//Go to the correct position in the level
	list< PositionTrace >::iterator PTi=(Li->traced_positions).begin();
	for(i=0;i<pos_index && PTi!=(Li->traced_positions).end();i++){
		PTi++;
	}
	
	//Update status
	PTi->status_unknown=false;	//false means "dead" (or "known status" if you're against violence)
}

//delete the lowest level
void Trace::delete_level() {
	QMutexLocker locker(&trace_lock);
	
	//"absolute" security to avoid any possible crash of pop_back
	if (traced_levels.empty()) {
		Error(QString("Trace::delete_level : there is no level to delete"), false);
		return;
	}
	
	if (current_level>0) { //"normal" security to avoid pop_back of an empty list
		current_level--;
		
		//if we went back to parent level, then parent has no more meaning (children trace)
		if(current_level<=parent_level) {
			parent_level=-1;
		}
		
		//delete the lowest level
		traced_levels.pop_back();
	}
}

//used to indicate a signal was received
void Trace::signal_received() {
	//no lock protection : it is always called "internally" from a protected function
	signal=0;
	reverse_zapping=false;
	signal_child_number=-1;
}

//return true if user asked to stop computation
bool Trace::stop_signal() {
	QMutexLocker locker(&trace_lock);
	
	//stop computation on signal stop_computing
	if (stop_computing) {
		return true;
	}
	
	return false;
}

//return true until the computation goes back to the correct level
//"for_cn" (cn means compute_nimber) allow to treat the case current_level==signal correctly,
//as this case is different in game::compute_nimber and game::compute_win_loss
bool Trace::user_interrupt(bool for_cn) {
	QMutexLocker locker(&trace_lock);
	
	//request is for above level (=> end this level)
	if ( (signal!=0 && current_level>=signal && !(current_level==signal && for_cn)) || stop_computing) {
		return true;
	}
	
	//invalid request (too low, requested level doesn't exist or zapping on this level has no meaning)
	//=>just receive the signal without any special treatment
	if(signal!=0) {
		signal_received();	//simply indicates signal was received
	}

	return false;
}

//detect any signal for zapping, and translate it into an integer treated in Game::compute_win_loss
int Trace::user_request() {
	QMutexLocker locker(&trace_lock);
	
	//request is for above levels (-> end this level)
	if ( (signal!=0 &&current_level>signal) || stop_computing ) {
		return (-1);			//means "go back to above level"
	}
	
	//request is for the current level (-> change the studied child)
	if(signal!=0 && current_level==signal) {
		bool rz_tmp=reverse_zapping;
		int scn_tmp=signal_child_number;
		signal_received();	//indicates signal was received
		if(scn_tmp!=-1){	//zap from the children's table
			return scn_tmp;		//means "go the indicated child"
		}else{			//go to next unknown position
			if (!rz_tmp){
				return -2;	//means "go to next child"
			} else {
				return -3;	//means "go to previous child"
			}
		}
	}
	
	//invalid request (too low, requested level doesn't exist)
	//->just receive the signal without any special treatment
	if(signal!=0 && current_level<signal) {
		signal_received();	//indicates signal was received
	}
	
	return (-4);				//means "do nothing"
}


//---------used in graphic thread ---------------

//number of positions in a given level
int Trace::level_size(int level_index) {
	//note : function called from an already locked point
	
	//find the correct level
	list< Level >::const_iterator Li=traced_levels.begin(); //level list iterator
	int i;
	for(i=1;i<=level_index;i++) {
		Li++;
	}
	
	//return the number of positions
	return (Li->traced_positions).size();
}

//set value of stop_computing
void Trace::set_stop_value(bool new_value) {
	QMutexLocker locker(&trace_lock);
	
	stop_computing=new_value;
}

//user interface->trace : user requested a specific child
void Trace::request_child(int index) {
	QMutexLocker locker(&trace_lock);
	
	//the second test avoids problems when clicking on children's table while those children are no longer studied
	if (signal==0 && parent_level!=-1) {
		signal=parent_level +1;
		signal_child_number=index;
	}
}

//user interface->trace : user requested next child on the specified level
void Trace::request_level(int index, bool backward) {
	QMutexLocker locker(&trace_lock);
	
	//note : class Game should always receive signals - even invalid
	//if a previous signal has not yet been received, there is no emission of a new one
	if (signal==0) {
		
		signal=index;	//emit the signal indicating which level should compute next branch (user choice)
		reverse_zapping=backward;
	}
}
