/***************************************************************************
 *   Copyright (C) 2006 by Julien Lemoine, Simon Viennot                   *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef TRACE_H
#define TRACE_H

#include <QMutex>
#include <QMutexLocker>

#include "position.h"

//Little class to store a position with special information useful when tracing
class PositionTrace {
  public :
	Line r_vtx;             //vertex representation
	nimber nim;             //nimber
	int nb_lives;           //number of lives
	bool status_unknown;    //whether status (W/L) is known or not
};

//A level is the set of children of a studied position
//You may see the computation state at a certain time, as the choice of a special children in each level
//Note that when you choose a different children in a certain level,
//the under-levels are destroyed and new ones are computed
class Level {
  public :
	list< PositionTrace > traced_positions;  //list of positions of this level
	int studied_pos_index;                   //index of studied position (used in Game::compute_win_loss), first one=0
	int nb_alive_pos;                        //number of still alive positions
	bool nimber_level;                       //true if this level is for a nimber 0-1-2 computing
	
	Level() {studied_pos_index=0;};
	~Level() {};
};


class WidgetTrace;
class WidgetChildren;

//main class to trace computation progress
//this class is thread-safe : all public functions (interface) are protected with a QMutex
class Trace {
  private:
	//----------Explanation of the values of signal/current_level/parent_level----------
	//The number of the first level, where are stored the children of the computed position, is 1.
	//The number of the second level is 2, etc. The number of a level is the same that the
	//number printed on the left of the table.
	//Be careful, row(item) is equal to "number_of_the_level(item)-1".
	//When we click on the level n for classical zapping, signal is set to n.
	//When we click on the level n to create a children table, parent_level is set to n,
	//and then when we click on a child in the children table, signal is set to n+1.
	//----------End of explanation----------
	
	int signal;                 //signal to change studied child on the specified level (0 if it has no meaning)
	int signal_child_number;    //signal for zapping from the children's table (first child=0, -1 if it has no meaning)
	bool stop_computing;        //for interruption because of an error, or by the player
	bool reverse_zapping;       //true if we are zapping back
	
	void add_one_pos_nim(const Pos_nim& pn); //add only one position-nimber to the bottom-level
	void signal_received();     //executed when signal is received
	
	//-------------objects used to store the data---------------------
	//WidgetTrace and WidgetChildren are friends because they need direct access to the data
	friend class WidgetTrace;
	friend class WidgetChildren;
	
	int current_level;	//current level in the game tree
	int parent_level;	//father's level of positions displayed in childrenWidget (-1 if it has non meaning)
	
	//database to trace the computation : list of currently computed levels
	list<Level> traced_levels;
	
  public:
	QMutex trace_lock;
	
	Trace() {};
	~Trace() {};
	
	//Interface (content)
	void initialize();                       //initialize the trace object (purge it)
	void create_level(bool for_nimber);      //create a new level
	void add(const Pos_nim& pn);             //add a position to the bottom-level
	void add(const list<Pos_nim> & l_pn);    //add a list of positions to the bottom-level
	void delete_level();                     //delete the bottom-level
	void change_studied_pos(int pos_index);  //compute->trace : indicate the studied position changed
	void known_child(int pos_index);         //compute->trace : indicate a child status was found
	int level_size(int level_index);         //number of positions of a given level
	
	//Interface (signals)
	bool stop_signal();                      //trace->compute : return true when user stops computation
	bool user_interrupt(bool for_cn);        //trace->compute : return true until the computation goes back to the correct level
	int user_request();                      //trace->compute : return int value describing the current user request
	void set_stop_value(bool value);         //user interface->trace : set value of stop_computing (used to stop a computation)
	void request_child(int index);                 //user interface->trace : user requested a specific child in the child list
	void request_level(int index, bool backward);  //user interface->trace : user requested next child on the specified level
};


//little class to create and destroy levels easily (it's the same concept as QMutexLocker)
class Level_trace {
  private:
	Trace* trace_object;
	
  public:
	//create a new level when object is created
	Level_trace(Trace *trace, bool for_nimber) {
		trace_object=trace;
		trace_object->create_level(for_nimber);
	};
	
	//delete level when object is destroyed
	~Level_trace() {
		trace_object->delete_level();
	};
};

#endif
